"""
Module for solving CommonRoad scenarios using trained models
"""
import argparse
import logging
import os

from commonroad_dc.feasibility.feasibility_checker import FeasibilityException

os.environ["KMP_WARNINGS"] = "off"
os.environ["KMP_AFFINITY"] = "none"
logging.getLogger("tensorflow").disabled = True
from typing import Union, List
import gym
import yaml
import warnings
import numpy as np
from commonroad.common.solution import (PlanningProblemSolution, Solution, CostFunction, CommonRoadSolutionWriter,
                                        TrajectoryType, )
from commonroad.scenario.scenario import ScenarioID, Scenario
from commonroad.common.solution import VehicleModel
from commonroad.planning.planning_problem import PlanningProblemSet
from commonroad.scenario.trajectory import Trajectory, State
from commonroad_dc.feasibility.solution_checker import valid_solution, _simulate_trajectory_if_input_vector, \
    GoalNotReachedException, CollisionException
from gym import Env

from commonroad_rl.utils_run.vec_env import CommonRoadVecEnv
from commonroad_rl.gym_commonroad.commonroad_env import CommonroadEnv
from commonroad_rl.gym_commonroad.constants import PATH_PARAMS
from commonroad_rl.utils_run.utils import load_model_and_vecnormalize


LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.DEBUG)
handler = logging.StreamHandler()
handler.setLevel(logging.DEBUG)
LOGGER.addHandler(handler)


def get_parser():
    parser = argparse.ArgumentParser(description="Evaluates PPO2 trained model with specified test scenarios",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument("--algo", type=str, default="ppo2")
    parser.add_argument("--test_path", "-i", type=str, help="Path to pickled test scenarios",
                        default=PATH_PARAMS["test_reset_config"], )
    parser.add_argument("--model_path", "-model", type=str, help="Path to trained model", required=True)
    parser.add_argument("--multiprocessing", "-mpi", action="store_true")
    parser.add_argument("--solution_path", "-sol", type=str,
                        help="Path to the desired directory of the generated solution files",
                        default=PATH_PARAMS["commonroad_solution"], )
    parser.add_argument("--cost_function", "-cost", type=str, default="JB1")
    parser.add_argument("--hyperparam_filename", "-f", type=str, default="model_hyperparameters.yml")
    parser.add_argument("--verbose", "-v", action="store_true")
    parser.add_argument("--debug", action="store_true", help="Debug mode (overrides verbose mode)")
    parser.add_argument("--config_filename", "-config_f", type=str, default="environment_configurations.yml", )

    return parser


def create_solution(commonroad_env: CommonroadEnv, output_directory: str, cost_function: str,
                    computation_time: Union[float, None] = None, processor_name: Union[str, None] = "auto",
                    input_trajectory: bool = True) -> bool:
    """
    Creates a CommonRoad solution file from the terminated environment

    :param commonroad_env: The terminated environment
    :param output_directory: The directory where the solution file will be written
    :param cost_function: The cost function to be used during the solution generation
    :param computation_time: The elapsed time during solving the scenario
    :param processor_name: The name of the used processor
    :return: True if the solution is valid and the solution file is written
    """
    os.makedirs(output_directory, exist_ok=True)
    list_state = list()
    ego_vehicle = commonroad_env.ego_action.vehicle
    for state in ego_vehicle.state_list:
        model_type = ego_vehicle.vehicle_model
        if model_type == VehicleModel.PM:
            if input_trajectory:
                if state.time_step == 0:
                    # skipping first accelerations since they are not executed
                    continue
                kwarg = {"time_step": state.time_step-1, # shift time step for input states
                         "acceleration": state.acceleration, "acceleration_y": state.acceleration_y}
            else:
                kwarg = {"position": state.position, "velocity": state.velocity, "velocity_y": state.velocity_y,
                         "time_step": state.time_step, "orientation": np.arctan2(state.velocity_y, state.velocity)}
        else:
            kwarg = {"position": state.position, "velocity": state.velocity, "steering_angle": state.steering_angle,
                     "orientation": state.orientation, "time_step": state.time_step}
        list_state.append(State(**kwarg))

    trajectory = Trajectory(initial_time_step=list_state[0].time_step, state_list=list_state)

    planning_problem_solution = PlanningProblemSolution(
            planning_problem_id=commonroad_env.planning_problem.planning_problem_id, vehicle_model=model_type,
            vehicle_type=ego_vehicle.vehicle_type, cost_function=CostFunction[cost_function], trajectory=trajectory, )
    scenario_id = ScenarioID.from_benchmark_id(commonroad_env.benchmark_id, "2020a")
    solution = Solution(scenario_id=scenario_id,  # TODO: wrong usage, fix in commonroad environment to use scenario_id
                        planning_problem_solutions=[planning_problem_solution], computation_time=computation_time,
                        processor_name=processor_name, )

    # # =================================== DEBUG ============================================================
    # planned_list = []
    # for state in ego_vehicle.state_list:
    #     kwargs = {"position": state.position, "velocity": state.velocity, "velocity_y": state.velocity_y,
    #               "time_step": state.time_step, "orientation": np.arctan2(state.velocity_y, state.velocity),
    #                  "acceleration": state.acceleration, "acceleration_y": state.acceleration_y}
    #     planned_list.append(State(**kwargs))
    # write_solution_file(commonroad_env.scenario, PlanningProblemSet([commonroad_env.planning_problem]),
    #                     solution, output_directory, planned_list=planned_list)
    # # ===================================  END ============================================================
    # Check the solution
    LOGGER.debug(f"Check solution of {commonroad_env.benchmark_id}")
    solution_valid, results = valid_solution(commonroad_env.scenario,
                                             PlanningProblemSet([commonroad_env.planning_problem]), solution, )
    if solution_valid:
        # write solution to a xml file
        csw = CommonRoadSolutionWriter(solution=solution)
        csw.write_to_file(output_path=output_directory, overwrite=True)
        LOGGER.info(f"Solution feasible, {commonroad_env.benchmark_id} printed to {output_directory}")
    else:
        LOGGER.info(f"Unable to create solution, invalid trajectory!")
    return solution_valid


def solve_scenarios(test_path: str, model_path: str, algo: str, solution_path: str, cost_function: str,
                    multiprocessing: bool = False, hyperparam_filename: str = "model_hyperparameters.yml",
                    config_filename: str = "environment_configurations.yml", env_id: str = "commonroad-v1",
                    render: bool = False, model_name="best_model"
                    ) -> List[bool]:
    """
    Solve a batch of scenarios using a trained model

    :param test_path: Path to the test files
    :param model_path: Path to the trained model
    :param algo: the used RL algorithm
    :param solution_path: Path to the folder where the solution files will be written
    :param cost_function: The cost function to be used during the solution generation
    :param multiprocessing: Indicates whether using multiprocessing or not (default is False)
    :param hyperparam_filename: The filename of the hyperparameters (default is model_hyperparameters.yml)
    :param config_filename: The environment configuration file name (default is environment_configurations.yml)
    :return: List of boolean values which indicates whether a scenario has been successfully solved or not
    """

    # mpi for parallel processing
    if multiprocessing:
        try:
            from mpi4py import MPI
        except ImportError:
            MPI = None
        if MPI is not None:
            rank = MPI.COMM_WORLD.Get_rank()
            test_path = os.path.join(test_path, str(rank))

    # Get environment keyword arguments including observation and reward configurations
    config_fn = os.path.join(model_path, config_filename)
    with open(config_fn, "r") as f:
        env_kwargs = yaml.load(f, Loader=yaml.Loader)

    env_kwargs["termination_configs"].update({"terminate_on_friction_violation": False})
    env_kwargs["goal_configs"].update({"relax_is_goal_reached": False})
    env_kwargs.update(
        {"meta_scenario_path": os.path.join(test_path, 'meta_scenario'),
         "test_reset_config_path": os.path.join(test_path, 'problem_test')}
    )

    def env_fn():
        return gym.make(env_id, play=True, **env_kwargs)

    env = CommonRoadVecEnv([env_fn])
    results = []

    def on_reset_callback(env: Union[Env, CommonroadEnv], elapsed_time: float):
        if env.observation_dict["is_goal_reached"][0] and not env.observation_dict["is_collision"][0]:
            LOGGER.info("Goal reached")
            os.makedirs(solution_path, exist_ok=True)
            solution_valid = create_solution(env, solution_path, cost_function, computation_time=elapsed_time)
        else:
            goal = env.planning_problem.goal
            state = env.ego_action.vehicle.state
            goal.is_reached(state)
            LOGGER.info("Goal not reached")
            LOGGER.info(f"Termination reason: {env.termination_reason}")
            solution_valid = False

        results.append(solution_valid)

    env.set_on_reset(on_reset_callback)

    # Load model hyperparameters:
    hyperparam_fn = os.path.join(model_path, hyperparam_filename)
    with open(hyperparam_fn, "r") as f:
        hyperparams = yaml.load(f, Loader=yaml.Loader)

    normalize = hyperparams["normalize"]

    model, env = load_model_and_vecnormalize(model_path, algo, normalize, env, model_name=model_name)

    obs = env.reset()
    if render:
        env.render()
    while True:
        action, _states = model.predict(obs, deterministic=True)
        try:
            obs, reward, done, info = env.step(action)
            if render:
                env.render()
            LOGGER.info(f"Step: {env.venv.envs[0].current_step}, \tReward: {reward}, \tDone: {done}")
            if info[0].get("out_of_scenarios", False):
                break
        except IndexError as e:
            # If the environment is done, it will be reset.
            # However the reset throws an exception if there are no more scenarios to be solved.
            LOGGER.info(f"Cannot choose more scenarios to be solved, msg: {e}")
            break
    return results


def write_solution_file(scenario: Scenario,
                        planning_problem_set: PlanningProblemSet,
                        solution: Solution, output_directory: str, planned_list: List[State]) -> bool:
    # print(f"Check validity of solution for {scenario.benchmark_id}")
    import matplotlib
    matplotlib.use("TkAgg")
    import matplotlib.pyplot as plt
    try:
        solution_valid, results = valid_solution(scenario, planning_problem_set, solution)
    except FeasibilityException as e:
        warnings.warn(str(e))
        planned_input_list = solution.planning_problem_solutions[0].trajectory.state_list
        # draw inputs
        plt.figure()
        plt.subplot(2, 1, 1)
        plt.plot(list(range(len(planned_input_list))),
                 [state.steering_angle_speed for state in planned_input_list], color="black", label="planned")
        plt.plot([0, len(planned_input_list)], [veh_params.steering.v_min, veh_params.steering.v_min],
                 color="red", label="bounds")
        plt.plot([0, len(planned_input_list)], [veh_params.steering.v_max, veh_params.steering.v_max],
                 color="red", label="bounds")
        plt.legend()
        plt.ylabel("steering angle velocity")
        plt.subplot(2, 1, 2)
        plt.plot(list(range(len(planned_input_list))),
                 [state.acceleration for state in planned_input_list], color="black", label="planned")
        plt.plot([0, len(planned_input_list)], [-veh_params.longitudinal.a_max, -veh_params.longitudinal.a_max],
                 color="red", label="bounds")
        plt.plot([0, len(planned_input_list)], [veh_params.longitudinal.a_max, veh_params.longitudinal.a_max],
                 color="red", label="bounds")
        plt.ylabel("acceleration")
        plt.show()

        return

    except (CollisionException, GoalNotReachedException) as e:
        warnings.warn(str(e))
        csw = CommonRoadSolutionWriter(solution=solution)
        csw.write_to_file(output_path=output_directory, overwrite=True)
        planning_problem_solution = solution.planning_problem_solutions[0]
        _, reconstructed_trajectory = _simulate_trajectory_if_input_vector(planning_problem_set,
                                                                           planning_problem_solution,
                                                                           scenario.dt)
        for state in reconstructed_trajectory.state_list:
            state.orientation = np.arctan2(state.velocity_y, state.velocity)

        # plt.figure(figsize=(20, 20))
        # # plt.subplot(2, 1, 1)
        # draw_object(scenario, plot_limits=[-50, 50, -50, 50])
        # # draw_object(planning_problem_set)
        # planned_state_list = planned_list
        # plt.plot(*np.array([state.position for state in planned_state_list]).T,
        #          color='black', marker='o', markersize=0.2, zorder=20, linewidth=0.5, label='planned trajectories')
        # plt.plot(*np.array([state.position for state in reconstructed_trajectory.state_list]).T,
        #          color='blue', marker='o', markersize=0.2, zorder=20, linewidth=0.5,
        #          label='reconstructed trajectories')
        # plt.legend()
        # plt.gca().set_aspect('equal')
        # plt.subplot(2, 1, 2)
        # plt.title("orientation")
        # plt.plot([state.orientation for state in planned_state_list], color="black")
        # plt.plot([state.orientation for state in reconstructed_trajectory.state_list], color="blue")

        # plt.savefig("trajectories.png", dpi=300)
        # plt.show()
        return

    if solution_valid:
        csw = CommonRoadSolutionWriter(solution=solution)
        csw.write_to_file(output_path=output_directory, overwrite=True)
    else:
        # draw planned and reconstructed inputs/trajectories
        planning_problem_solution = solution.planning_problem_solutions[0]
        if not planning_problem_solution.trajectory_type is TrajectoryType.Input:
            # draw inputs
            warnings.warn("Planned inputs and reconstructed inputs deviate too much!")
            planned_input_list = planned_list
            reconstructed_input_list = results[1][1].state_list
            plt.figure()
            plt.subplot(2, 1, 1)
            plt.plot(list(range(len(planned_input_list))),
                     [state.steering_angle_speed for state in planned_input_list], color="black", label="planned")
            plt.plot(list(range(len(reconstructed_input_list))),
                     [state.steering_angle_speed for state in reconstructed_input_list], color="blue",
                     label="reconstructed")
            plt.legend()
            plt.ylabel("steering angle velocity")
            plt.subplot(2, 1, 2)
            plt.plot(list(range(len(planned_input_list))),
                     [state.acceleration for state in planned_input_list], color="black", label="planned")
            plt.plot(list(range(len(reconstructed_input_list))),
                     [state.acceleration for state in reconstructed_input_list], color="blue", label="reconstructed")
            plt.ylabel("acceleration")
            plt.show()


def load_and_check_solution(scenario_file_path: str, solution_file_path: str) -> bool:
    """
    Function to check whether a solution file is valid for a scenario or not

    :param scenario_file_path: Path to the scenario
    :param solution_file_path: Path to the solution
    :return: True if the solution is valid
    """
    from commonroad.common.file_reader import CommonRoadFileReader
    from commonroad.common.solution import CommonRoadSolutionReader

    scenario, pp = CommonRoadFileReader(scenario_file_path).open()
    solution = CommonRoadSolutionReader.open(solution_file_path)
    solution_valid, results = valid_solution(scenario, pp, solution)
    if solution_valid is True:
        LOGGER.info(f"Solution feasible")
    else:
        LOGGER.info(f"Unable to create solution, invalid trajectory!")
    return solution_valid


if __name__ == "__main__":
    args = get_parser().parse_args()

    if args.verbose:
        LOGGER.setLevel(logging.INFO)
    if args.debug:
        LOGGER.setLevel(logging.DEBUG)

    solve_scenarios(args.test_path, args.model_path, args.algo, args.solution_path, args.cost_function,
                    multiprocessing=args.multiprocessing, hyperparam_filename=args.hyperparam_filename,
                    config_filename=args.config_filename, )



import yaml
import numpy as np
import matplotlib.pyplot as plt
from abc import ABC
from collections import defaultdict
from typing import List

from shapely.geometry import LineString
from shapely.geometry import Polygon as sha_Polygon
from shapely.ops import unary_union
from scipy import spatial
from scipy.interpolate import splprep, splev

from commonroad.scenario.trajectory import State
from commonroad.scenario.scenario import Scenario
from commonroad.scenario.lanelet import Lanelet
from commonroad.geometry.shape import Polygon as Polygon

import commonroad_dc.pycrcc as pycrcc
# from commonroad_dc.geometry.geometry import CurvilinearCoordinateSystem
from commonroad_dc.pycrccosy import CurvilinearCoordinateSystem
from commonroad_dc.geometry.util import compute_curvature_from_polyline, compute_orientation_from_polyline, \
    compute_pathlength_from_polyline, resample_polyline
from commonroad_dc.collision.collision_detection.pycrcc_collision_dispatch import create_collision_object
from commonroad_rl.gym_commonroad.constants import PATH_PARAMS
from commonroad_rl.gym_commonroad.utils.scenario import angle_difference, approx_orientation_vector


__author__ = "Yinqiang Zhang, Xun Wang"
__copyright__ = "TUM Cyber-Physical System Group"
__credits__ = [""]
__version__ = "0.1"
__email__ = "yinqiang.zhang@tum.de"
__status__ = "Development"


class ConflictZone(ABC):
    """
    Description:
        Class for conflict zones constructed by intersection areas

    important concepts and variables:
    1. intersection-based conflict zones:
    2. lane-based conflict zones:
    3. unsafe lane pair pattern (groups):
    """

    def __init__(self, reaching_time_prediction: bool =True):
        """
        initializes basic variables of conflict zones

        :param reaching_time_prediction: chooses which method is used to predict reaching time of other vehicles.
        """
        self.scenario = None
        self.lane_list = None  # lane list generate by reference manager
        self.intersection_conflict_zone = None  # intersection coflict region: C_I
        self.navigator = None

        # lane-based conflict zone
        self.lane_conflict_information = dict()
        # unsafe lane pair
        self.unsafe_lane_pair_pattern = dict()
        # collision model for intersection conflict zone
        self.region_collision_model = None
        # variable cache for computation acceleration
        self.cache_conflict_zone = defaultdict(lambda: None)

        # safe distance definition
        self.delta_conflict_region_width = 1.5
        self.reaching_time_based_prediction = reaching_time_prediction
        self.config_file = PATH_PARAMS["configs"]["commonroad-v1"]
        with open(self.config_file, "r") as root_path:
            self.config = yaml.safe_load(root_path)
        # TODO: now hardcoded - if necessary load from config file or vehicle parameter
        #self.maximal_velocity = self.config['occupany_predictor']['maximal_velocity']
        self.maximal_velocity = 20.0
        self.weight_coefficient = 8 #TODO: necessary to initialize here?

    def reset(self, scenario: Scenario):
        """
        resets the class with new scenario

        :param scenario: base scenario used in planning
        """
        self.scenario = scenario
        identifer = '_'.join(scenario.scenario_id.__str__().split('_')[0:2])

        if self.cache_conflict_zone[identifer] is not None:
            self.lane_list = self.cache_conflict_zone[identifer]['lane_list']
            self.intersection_conflict_zone = self.cache_conflict_zone[identifer]['intersection_conflict_zone']
            self.lane_conflict_information = self.cache_conflict_zone[identifer]['conflict_information']
            self.unsafe_lane_pair_pattern = self.cache_conflict_zone[identifer]['unsafe_pattern']
            self.lane_velocity_upper_bound = self.cache_conflict_zone[identifer]['velocity_bound']
            self.lane_min_time_interval = self.cache_conflict_zone[identifer]['time_interval']
        else:
            self.lane_list = self._create_lanes_for_lanelet_network()
            self.intersection_conflict_zone = self._conflict_region_generation(threshold_area=30)
            self.lane_conflict_information = self._generate_conflict_information()
            self.unsafe_lane_pair_pattern = self._generate_unsafe_pattern()
            self.lane_velocity_upper_bound = self.get_curvature_velocity_constraints(a_lat_max=4.0, v_max=20.0)
            self.lane_min_time_interval = self.get_minimal_time_to_drive()
            # save in cache
            temp_dict = dict()
            temp_dict['lane_list'] = self.lane_list
            temp_dict['intersection_conflict_zone'] = self.intersection_conflict_zone
            temp_dict['conflict_information'] = self.lane_conflict_information
            temp_dict['unsafe_pattern'] = self.unsafe_lane_pair_pattern
            temp_dict['velocity_bound'] = self.lane_velocity_upper_bound
            temp_dict['time_interval'] = self.lane_min_time_interval
            self.cache_conflict_zone[identifer] = temp_dict

        # generate collsion model for intersection conflict zone
        self.conflict_collision_model = list()
        for region in self.intersection_conflict_zone:
            self.conflict_collision_model.append(self._generate_commonroad_collision_model(region))

    def _conflict_region_generation(self, threshold_area: float = 10):
        """
        generates intersection conflict region according to the current scenario
        :param threshold_area: the minimal area considered when generating the conflict zone
        :return:
            list of valid intersection-based conflict zones
        """
        show_list = list()
        conflict_regions = list()
        # only calculate once
        for ind, lane in enumerate(self.lane_list):
            target_lane = lane
            target_ids = lane[0]

            # compare with other lane
            for k in range(ind + 1, len(self.lane_list)):
                compared_lane = self.lane_list[k]
                compared_ids = compared_lane[0]

                # remove all the same lanelets
                trimmed_ids = list()
                for target_id in target_ids:
                    if target_id in compared_ids:
                        trimmed_ids.append(target_id)
                target_ids = list(set(target_ids).difference(set(trimmed_ids)))
                compared_ids = list(set(compared_ids).difference(set(trimmed_ids)))
                # no adjacent, but collision, not with same predecessor (3 conditions)
                # then calculate intersection area
                for target_id in target_ids:
                    target_lanelet = self.scenario.lanelet_network.find_lanelet_by_id(target_id)
                    target_polygon = target_lanelet.convert_to_polygon()
                    polygon1 = sha_Polygon(target_polygon.vertices)
                    for compared_id in compared_ids:
                        compared_lanelet = self.scenario.lanelet_network.find_lanelet_by_id(compared_id)
                        compared_polygon = compared_lanelet.convert_to_polygon()
                        polygon2 = sha_Polygon(compared_polygon.vertices)
                        if polygon1.overlaps(polygon2):
                            # recursively find the predecessors
                            from_same_lane = False
                            target_temp = target_lanelet
                            compared_temp = compared_lanelet
                            if target_temp.predecessor == compared_temp.predecessor:
                                from_same_lane = True
                            else:
                                while target_temp.predecessor != compared_temp.predecessor:
                                    # assume that if has different predecessors, only have one
                                    target_temp = self.scenario.lanelet_network.find_lanelet_by_id(
                                        target_temp.predecessor[0])
                                    compared_temp = self.scenario.lanelet_network.find_lanelet_by_id(
                                        compared_temp.predecessor[0])
                                    if len(target_temp.predecessor) == 0 or len(compared_temp.predecessor) == 0:
                                        break
                                    elif target_temp.predecessor == compared_temp.predecessor:
                                        from_same_lane = True
                                        break

                            if (not from_same_lane
                                    and ((target_id, compared_id) not in show_list
                                         and (compared_id, target_id) not in show_list)):
                                show_list.append((target_id, compared_id))
                                inter = polygon1.intersection(polygon2)
                                conflict_regions.append(inter)

        # calculate union
        conflict_union = unary_union(conflict_regions)
        try:
            valid_conflict_regions = list(region for region in conflict_union.geoms if region.area > threshold_area)
        except AttributeError:
            valid_conflict_regions = [conflict_union]

        # open and close operations to smooth the region
        for k in range(len(valid_conflict_regions)):
            valid_conflict_regions[k] = valid_conflict_regions[k].buffer(1.0)
            valid_conflict_regions[k] = valid_conflict_regions[k].buffer(-2.0)
            valid_conflict_regions[k] = valid_conflict_regions[k].buffer(1.0)

        return valid_conflict_regions

    def _generate_conflict_information(self):
        """
        generates conflict information for each lane

        :return:
            information tuple (lane information, conflict zone, collision model, critical points(curvi))
        """
        assert self.intersection_conflict_zone is not None, "please generate the conflict region first"

        conflict_information_dict = dict()
        for lane in self.lane_list:
            # values for different conflict region
            lane_conflict_regions = list()
            conflict_collision_models = list()
            lane_critical_points = list()
            lane_id = lane[1].lanelet_id
            shapely_lane_polygon = sha_Polygon(lane[1].convert_to_polygon().vertices).buffer(-0.1)
            for conflict_zone in self.intersection_conflict_zone:
                lane_conflict_region = conflict_zone.intersection(shapely_lane_polygon)

                lane_conflict_region = lane_conflict_region.buffer(-0.2)
                lane_conflict_region = lane_conflict_region.buffer(0.2)

                if lane_conflict_region.area != 0:
                    # calculate critical points
                    reference_line_string = LineString(lane[3].reference())
                    extended_conflict_region = lane_conflict_region.buffer(2)
                    line_segments = reference_line_string.intersection(extended_conflict_region)
                    before_critical_point = line_segments.coords[0]
                    after_critical_point = line_segments.coords[-1]
                    try:
                        before_s, _ = lane[3].convert_to_curvilinear_coords(before_critical_point[0],
                                                                            before_critical_point[1])
                    except ValueError:
                        # project the results to the center line
                        idx = np.argmin(np.linalg.norm(lane[3].reference() - before_critical_point, axis=1))
                        pos = lane[3].reference()[idx]
                        before_s, _ = lane[3].convert_to_curvilinear_coords(pos[0], pos[1])

                    try:
                        end_s, _ = lane[3].convert_to_curvilinear_coords(after_critical_point[0],
                                                                         after_critical_point[1])
                    except ValueError:
                        # project the results to the center line
                        idx = np.argmin(np.linalg.norm(lane[3].reference() - after_critical_point, axis=1))
                        pos = lane[3].reference()[idx]
                        end_s, _ = lane[3].convert_to_curvilinear_coords(pos[0], pos[1])

                    critical_points_pair = (before_s, end_s)
                    # generate collision model (expended)
                    conflict_collision_model = self._generate_commonroad_collision_model(lane_conflict_region)
                    lane_conflict_regions.append(lane_conflict_region)
                    conflict_collision_models.append(conflict_collision_model)
                    lane_critical_points.append(critical_points_pair)

            conflict_information_dict[lane_id] = (
            lane, lane_conflict_regions, conflict_collision_models, lane_critical_points)

        return conflict_information_dict

    def _generate_unsafe_pattern(self):
        """
        generate all unsafe lane pairs

        :return:
            unsafe pair dictionary
        """
        unsafe_pattern_dict = dict()
        for lane in self.lane_list:
            lane_id = lane[1].lanelet_id
            unsafe_pattern_dict[lane_id] = self._get_unsafe_lane_pair(lane)

        return unsafe_pattern_dict

    def _get_unsafe_lane_pair(self, ref_lane):
        """
        generates all unsafe pairs for the target lane

        :param reference lane: reference lane of ego vehicle
        :return:
            unsafe lane pairs
        """
        unsafe_pairs = defaultdict(list)

        ref_lane_id = ref_lane[1].lanelet_id
        ref_lane_conflict_information = self.lane_conflict_information[ref_lane_id]
        ref_lane_conlict_zone = ref_lane_conflict_information[1]
        ref_lane_sublanelet_ids = ref_lane[0]

        for lane_id in self.lane_conflict_information:
            lane_conflict_information = self.lane_conflict_information[lane_id]
            lane = lane_conflict_information[0]
            lane_sublanelet_ids = lane_conflict_information[0][0]
            lane_conflict_zone = lane_conflict_information[1]
            # not from the same direction
            if ref_lane_sublanelet_ids[0] != lane_sublanelet_ids[0]:
                for conflict_zone in lane_conflict_zone:
                    for ref_conlict_zone in ref_lane_conlict_zone:
                        cross_zone = ref_conlict_zone.intersection(conflict_zone)
                        if cross_zone.area != 0:
                            cross_point = cross_zone.centroid.coords[0]
                            # generate curvi position for conflict lanes
                            ref_lane_s, _ = ref_lane[3].convert_to_curvilinear_coords(cross_point[0], cross_point[1])
                            lane_s, _ = lane[3].convert_to_curvilinear_coords(cross_point[0], cross_point[1])
                            cross_commonroad_zone = self._generate_commonroad_collision_model(cross_zone)
                            unsafe_pairs[lane_id].append((ref_lane_s, lane_s, cross_commonroad_zone))

        return unsafe_pairs

    def get_unsafe_obstacle_association(self, conflict_obstacles_information):
        """
        generates the associations between dynamic obstacles and unsafe lane pairs

        :param conflict_obstacles_information: states of surrounding vehicles near intersection areas
        :return:
            unsafe obstacle-lane association
            intersection-related observations (states)
        """
        # TODO: currently not used - maybe delete later
        unsafe_association = list()
        observations = list()
        for state in conflict_obstacles_information:
            position = state.position
            orientation = state.orientation
            [lanlet_id_candidates] = self.scenario.lanelet_network.find_lanelet_by_position([position])

            delta_s = 50
            if len(lanlet_id_candidates) == 1:
                lanelet = self.scenario.lanelet_network.find_lanelet_by_id(lanlet_id_candidates[0])
                # not in the opposite direction of the lane
                if abs(self.get_lane_relative_heading(state, lanelet)) < 0.5 * np.pi:  # 90 degree
                    unsafe_information = self._get_lane_conflict_information(lanlet_id_candidates[0], state)
                    if len(unsafe_information) != 0:
                        temp_delta_s = list(unsafe_information.values())[0][1]
                        delta_s = temp_delta_s if temp_delta_s < delta_s else delta_s
                    unsafe_association.append((state, unsafe_information))
            else:
                for lanelet_id in lanlet_id_candidates:
                    # a threshold for relative heading
                    lanelet = self.scenario.lanelet_network.find_lanelet_by_id(lanelet_id)
                    if abs(self.get_lane_relative_heading(state, lanelet)) < 0.5:
                        unsafe_information = self._get_lane_conflict_information(lanelet_id, state)
                        if len(unsafe_information) != 0:
                            temp_delta_s = list(unsafe_information.values())[0][1]
                            delta_s = temp_delta_s if temp_delta_s < delta_s else delta_s
                        unsafe_association.append((state, unsafe_information))

            if delta_s != 50.0:
                observations.append((delta_s, state.velocity))

        # process observations
        observations = sorted(observations, key=lambda x: x[0])

        return unsafe_association, observations

    def generate_intersection_observation(self, conflict_obstacles_information):
        """
        generates the associations between dynamic obstacles and unsafe lane pairs

        :param conflict_obstacles_information: states of surrounding vehicles near intersection areas
        :return:
            intersection-related observations for other vehicles (distances and velocities)
        """
        observations = list()
        for state in conflict_obstacles_information:
            position = state.position
            [lanlet_id_candidates] = self.scenario.lanelet_network.find_lanelet_by_position([position])

            delta_s = 50
            if len(lanlet_id_candidates) == 1:
                lanelet = self.scenario.lanelet_network.find_lanelet_by_id(lanlet_id_candidates[0])
                # not in the opposite direction of the lane
                if abs(self.get_lane_relative_heading(state, lanelet)) < 0.5 * np.pi:  # 90 degree
                    unsafe_information = self._get_lane_conflict_information(lanlet_id_candidates[0], state)
                    if len(unsafe_information) != 0:
                        temp_delta_s = list(unsafe_information.values())[0][1]
                        delta_s = temp_delta_s if temp_delta_s < delta_s else delta_s
            else:
                for lanelet_id in lanlet_id_candidates:
                    # a threshold for relative heading
                    lanelet = self.scenario.lanelet_network.find_lanelet_by_id(lanelet_id)
                    if abs(self.get_lane_relative_heading(state, lanelet)) < 0.5:
                        unsafe_information = self._get_lane_conflict_information(lanelet_id, state)
                        if len(unsafe_information) != 0:
                            temp_delta_s = list(unsafe_information.values())[0][1]
                            delta_s = temp_delta_s if temp_delta_s < delta_s else delta_s

            if delta_s != 50.0:
                observations.append((delta_s, state.velocity))

        # process observations
        observations = sorted(observations, key=lambda x: x[0])

        return observations

    def check_intersection_safety(self, ego_state, ref_lane, unsafe_association, obs_a_stop, time_step, plan_horizon):
        """
        checks how long time the state of ego vehicle can satisfy the safety of intersection
        :param ego_state: the current state of ego vehicle
        :param ref_lane: current reference lane of ego vehicle
        :param unsafe_association: the vehicle-lane pairs that have a potential danger with ego vehicle
        :param obs_a_stop: largest acceleration used by obstacle before the intersection
        :param time_step: current time step
        :param time_step: planning horizon of low-level planner

        :return:
            time-variant collision model of conflict zones
        """
        # TODO: currently not used - maybe delete later
        # get information of ego reference lane and ego vehicle
        ref_lane_id = ref_lane[1].lanelet_id  # reference lane id
        unsafe_pattern = self.unsafe_lane_pair_pattern[ref_lane_id]  # all unsafe lane pairs
        conflict_information = self.lane_conflict_information[ref_lane_id]  # lane-based conflict information

        if len(conflict_information[3]) == 0:
            # lane without passing intersection area
            return pycrcc.ShapeGroup()
        else:
            # critical points before and after intersection
            ego_critical_near, ego_criical_far = conflict_information[3][0]
            # current longitudinal position of ego vehicle
            try:
                ego_s, _ = ref_lane[3].convert_to_curvilinear_coords(ego_state.position[0], ego_state.position[1])
            except ValueError:
                # project the results to the center line
                idx = np.argmin(np.linalg.norm(ref_lane[3].reference() - ego_state.position, axis=1))
                pos = ref_lane[3].reference()[idx]
                ego_s, _ = ref_lane[3].convert_to_curvilinear_coords(pos[0], pos[1])
            # safety verification converted into collision avoidance
            collision_object_list = pycrcc.ShapeGroup()
            if ego_s < ego_criical_far:  # ego vehicle not pass the intersection
                predicted_reach_time = 100  # 100 is a large enough number as a place holder
                unsafe_intersection = False
                for lane_id in unsafe_pattern:  # seacrch in all unsafe lane pairs
                    for state, lane_dict in unsafe_association:
                        if lane_dict.get(lane_id, None) is not None:
                            # obs_s: location of obstacle in its lane
                            # delta_obs_s: distance to the conflict region
                            obs_s, delta_obs_s = lane_dict.get(lane_id, None)
                            # the position of crossing point in lanes of one unsafe lane pair
                            ego_critcal_s, obs_critical_s, cross_collision_object = unsafe_pattern.get(lane_id)[0]
                            # distance between obstacle and its corresponding crossing point
                            safe_delta_distance = obs_critical_s - obs_s
                            # reaching_time_based_prediction

                            if self.reaching_time_based_prediction:
                                # method 1: constant-velocity prediction
                                if safe_delta_distance > self.delta_conflict_region_width:
                                    reach_time = self.get_minimum_time_to_conflict_region(lane_id, state, obs_s,
                                                                                          obs_critical_s)
                                    predicted_reach_time = reach_time if reach_time < predicted_reach_time else predicted_reach_time
                                elif abs(safe_delta_distance) <= self.delta_conflict_region_width:
                                    unsafe_intersection = True
                            else:
                                # method 2: safe braking prediction
                                if safe_delta_distance > self.delta_conflict_region_width:
                                    v_max = np.sqrt(2 * (obs_critical_s - obs_s) * obs_a_stop)
                                    if v_max <= state.velocity:
                                        v_0 = state.velocity
                                        reach_time = (v_0 - np.sqrt(v_0 ** 2 - v_max ** 2)) / obs_a_stop
                                        predicted_reach_time = reach_time if reach_time < predicted_reach_time else predicted_reach_time
                                elif abs(safe_delta_distance) <= self.delta_conflict_region_width:
                                    unsafe_intersection = True

                # print(unsafe_intersection, predicted_reach_time)
                return self._generate_conflict_collision_model(conflict_information[2][0],
                                                               unsafe_intersection, time_step, predicted_reach_time,
                                                               plan_horizon)
            else:
                # ego vehicle has passed the intersection
                return pycrcc.ShapeGroup()

    def _get_lane_conflict_information(self, sub_lanelet_id, state):
        """
        gets the corresponding usafe conflict information at intersection area
        param: sub_lanelet_id: the lanelet id of the obstacle
        param: state: the state of the obstacle

        :return:
            features of unsafe lane pair group
        """

        # find possible lanes
        lane_id_list = list()
        for lane in self.lane_list:
            if sub_lanelet_id in lane[0]:
                lane_id_list.append(lane[1].lanelet_id)

        # associate the obstacle with the corresponding lane
        unsafe_information = dict()
        for lane_id in lane_id_list:
            conflict_information = self.lane_conflict_information[lane_id]
            if len(conflict_information[3]) == 0:
                continue
            lane_information = conflict_information[0]
            try:
                obs_s, _ = lane_information[3].convert_to_curvilinear_coords(state.position[0], state.position[1])
            except ValueError:
                # project the results to the center line
                idx = np.argmin(np.linalg.norm(lane_information[3].reference() - state.position, axis=1))
                pos = lane_information[3].reference()[idx]
                obs_s, _ = lane_information[3].convert_to_curvilinear_coords(pos[0], pos[1])

            # get maixmal velocity for turning
            # contraints for predicted trajectory
            maximal_lateral_acceleration = 4  # a_lateral

            s_idx = np.argmin(np.abs(lane_information[3].ref_pos() - obs_s))
            if lane_information[3].ref_pos()[s_idx] > obs_s:
                s_idx -= 1
            s_lambda = (obs_s - lane_information[3].ref_pos()[s_idx]) / (
                    lane_information[3].ref_pos()[s_idx + 1] - lane_information[3].ref_pos()[s_idx])
            ego_curvature = (s_lambda) * lane_information[3].ref_curv()[s_idx + 1] + (1 - s_lambda) * \
                            lane_information[3].ref_curv()[s_idx]
            v_max_limit = min(self.maximal_velocity, np.sqrt(maximal_lateral_acceleration / np.abs(ego_curvature)))

            critical_point_near, critical_point_far = conflict_information[3][0]
            # obstacle before or in ther intersection and its velocity fits for the corresponding lane
            if obs_s < critical_point_far and v_max_limit > state.velocity:
                # longitudinal position of obstacle on current lane
                # distance to the conflict region
                unsafe_information[lane_id] = (obs_s, critical_point_near - obs_s)

        return unsafe_information

    def _generate_commonroad_collision_model(self, shapely_polygon):
        """
        convert the polygon in shaeply into the commonroad

        :return:
            collision model (commonroad)
        """
        points = shapely_polygon.exterior
        vertices = np.stack((np.array(points.xy[0].tolist()), np.array(points.xy[1].tolist())), axis=0).T
        # generate collision object
        return create_collision_object(Polygon(vertices))

    def _generate_conflict_collision_model(self, polygon_collision_object, unsafe_intersection, time_step, ttr,
                                           plan_horizon):

        """
        generates time-variant collision model for conflict zones
        :param polygon_collision_object: collision object (shape) of lane-based conflict zones
        :param unsafe_intersection: whether lane-based conflict zone is safe or not
        :param time_step: current time step
        :param ttr: time to reach the conflict zones
        :param plan_horizon: time horizon of reactive planner

        :return:
            time-variant collision object
        """
        if unsafe_intersection:
            collision_object = polygon_collision_object
        else:
            if ttr < 4:
                ttr_time_steps = int(ttr / self.scenario.dt)
                horizon_time_steps = int(plan_horizon / self.scenario.dt)

                collision_object = pycrcc.TimeVariantCollisionObject(time_step + ttr_time_steps)
                i = ttr_time_steps
                while i <= horizon_time_steps:
                    collision_object.append_obstacle(polygon_collision_object)
                    i += 1
            else:
                collision_object = pycrcc.ShapeGroup()

        return collision_object

    def _generate_collision_object(self):
        """
        generates objects for later collision checker
        """
        # TODO: consider multiple conflict regions
        # TODO: currently not used - maybe delete later
        points = self.intersection_conflict_zone.exterior
        vertices = np.stack((np.array(points.xy[0].tolist()), np.array(points.xy[1].tolist())), axis=0).T
        # generate collision object
        self.region_collision_model = create_collision_object(Polygon(vertices))
        self.region_velocity_bound = [0, 0]

    def draw_conflict_region(self, conflict_region=None, linewidth=1, color='chocolate'):
        """
        draws the generated conflict region in the render function
        """
        # TODO: integrate in surrounding observation draw function
        if conflict_region is None:
            assert self.intersection_conflict_zone is not None, "conflict region has not been generated yet"
            conflict_region = self.intersection_conflict_zone
        try:
            for item in conflict_region:
                plt.plot(*item.exterior.xy, zorder=21, linewidth=linewidth, color=color)
        except:
            plt.plot(*conflict_region.exterior.xy, linewidth=linewidth, zorder=21, color=color)

    def check_near_intersection_area(self, state: State, ref_lane, distane_threshold):
        """
        checks if the ego vehicle is near enough to the intersection areas

        :param state: current state of ego vehicle
        :param ref_lane: reference lane of ego vehicle
        :param distane_threshold: distance to the intersection area

        :return:
            if lane change is feasible
            distance to the nearest boundary of conflict zone
            distance to the farest boundary of conflict zone
        """
        # TODO: currently not used - maybe delete later
        ref_lane_id = ref_lane[1].lanelet_id
        ref_lane_conflict_information = self.lane_conflict_information[ref_lane_id]
        if len(ref_lane_conflict_information[3]) == 0:
            return True, -10, -0.1  # set feasible to change the lane
        critical_point_near, critical_point_far = ref_lane_conflict_information[3][0]

        # calculate current position in curvilinear system
        position = state.position
        ego_s, _ = ref_lane[3].convert_to_curvilinear_coords(position[0], position[1])

        delta_s_near, delta_s_far = critical_point_near - ego_s, critical_point_far - ego_s

        lane_changing_feasible = True
        if delta_s_near < distane_threshold and delta_s_far > 0:
            lane_changing_feasible = False

        return lane_changing_feasible, delta_s_near, delta_s_far

    def get_ego_intersection_observation(self, position: np.array):
        """
        checks if the ego vehicle is near enough to the intersection areas

        :param position: current position of ego vehicle

        :return:
            distance to the nearest boundary of conflict zone
            distance to the farest boundary of conflict zone
        """
        ego_lane_id = self.scenario.lanelet_network.find_lanelet_by_position([position])[0][0]
        for lane in self.lane_list:
            if ego_lane_id in lane[0]:
                ref_lane_id = lane[1].lanelet_id
                break
        ref_lane_conflict_information = self.lane_conflict_information[ref_lane_id]
        if len(ref_lane_conflict_information[3]) == 0:
            return -10, -0.1  # set feasible to change the lane
        critical_point_near, critical_point_far = ref_lane_conflict_information[3][0]

        # calculate current position in curvilinear system

        cosy = None
        for lane in self.lane_list:
            if ego_lane_id in lane[0]:
                cosy = lane[3]
                break
        if cosy is None:
            print('ERROR:ConflictZone: coordinate system for calculating intersection ego observation not found')
            raise ValueError
        try:
            ego_s, _ = cosy.convert_to_curvilinear_coords(position[0], position[1])
        except ValueError:
            # project the results to the center line
            idx = np.argmin(np.linalg.norm(cosy.reference() - position, axis=1))
            pos = cosy.reference()[idx]
            ego_s, _ = cosy.convert_to_curvilinear_coords(pos[0], pos[1])
            
        delta_s_near, delta_s_far = critical_point_near - ego_s, critical_point_far - ego_s

        return delta_s_near, delta_s_far

    def get_curvature_velocity_constraints(self, a_lat_max, v_max=None):
        """
        gets the velocity constraints of lane caused by curvature
        :param lane: the target lane
        :param a_lat_max: maximal lateral acceleration
        :param v_max: speed limit of the current lane

        :return:
            upper bound of velocity
        """
        lane_velocity_upper_bound = defaultdict(list)
        for lane in self.lane_list:
            lane_id = lane[1].lanelet_id
            ref_curvature = lane[3].ref_curv()
            max_velocity = self.maximal_velocity if v_max is None else v_max
            for curv in ref_curvature:
                v_max = min(max_velocity, np.sqrt(a_lat_max / max(curv, 0.005)))
                lane_velocity_upper_bound[lane_id].append(v_max)
        return lane_velocity_upper_bound

    def get_minimal_time_to_drive(self):
        """
        gets minimal time to traverse

        return:
            minimal reaching time
        """
        min_time_interval = defaultdict(list)
        for lane in self.lane_list:
            lane_id = lane[1].lanelet_id
            ref_position = lane[3].ref_pos()
            velocity_bound = self.lane_velocity_upper_bound[lane_id]
            for k in range(len(velocity_bound) - 1):
                delta_s = ref_position[k + 1] - ref_position[k]
                # 1. the most conservative
                # v_max = max(velocity_bound[k+1], velocity_bound[k])
                # 2. less conservative
                v_max = (velocity_bound[k + 1] + velocity_bound[k]) / 2
                delta_t_min = delta_s / v_max
                min_time_interval[lane_id].append(delta_t_min)

        return min_time_interval

    def get_minimum_time_to_conflict_region(self, lane_id, state, obs_s, obs_critical_s):

        """
        gets the reaching time for constant-velocity-based approach

        :param lane_id: ID of current lanelet
        :param state: current state of ego vehicle
        :param obs_s: longitudinal position of obstacle
        :param obs_critical_s: longitudinal position of crossing point

        :return:
            predicted reaching time
        """

        # lane where obstacle is driving
        conflict_information = self.lane_conflict_information[lane_id]
        lane_information = conflict_information[0]

        ref_position = lane_information[3].ref_pos()
        delta_t_list = self.lane_min_time_interval[lane_id]
        velocity_bound = self.lane_velocity_upper_bound[lane_id]

        change_s_idx = np.where(velocity_bound <= state.velocity)[0]
        if len(change_s_idx) == 0:
            # constant velocity assumption
            time_to_conflict = (obs_critical_s - obs_s) / max(0.001, state.velocity)
        else:
            begin_s_idx = change_s_idx[0]
            t_constant = (ref_position[begin_s_idx] - obs_s) / max(0.001, state.velocity)
            s_idx_conflict = np.argmin(np.abs(ref_position - obs_critical_s))
            time_to_conflict = t_constant + np.sum(delta_t_list[begin_s_idx:s_idx_conflict])

        return time_to_conflict

    def check_in_conflict_region(self, ego_vehicle):
        """
        checks if the ego vehicle is in the intersection area now

        :return:
            whether the ego vehicle is in the intersection area
        """
        # TODO: currently not used - maybe delete later
        in_intersection = False
        for cc in self.conflict_collision_model:
            if cc.collide(ego_vehicle.collision_object):
                in_intersection = True
                break
        return in_intersection

    def get_lane_relative_heading(self, ego_vehicle_state: State, ego_vehicle_lanelet: Lanelet) -> float:
        """
        Get the heading angle in the Frenet frame.

        :param ego_vehicle_state: state of ego vehicle
        :param ego_vehicle_lanelet: lanelet of ego vehicle
        :return: heading angle in frenet coordinate system relative to lanelet center vertices between -pi and pi
        """
        # TODO: This function is a duplicate with the function in EgoObservation
        lanelet_angle = self._get_orientation_of_polyline(ego_vehicle_state.position,
                                                                    ego_vehicle_lanelet.center_vertices)

        return angle_difference(approx_orientation_vector(lanelet_angle),
                                approx_orientation_vector(ego_vehicle_state.orientation))

    @staticmethod
    def _get_orientation_of_polyline(position: np.array, polyline: np.array) -> float:
        """
        Get the approximate orientation of the lanelet.

        :param position: position to calculate the orientation of lanelet
        :param polyline: polyline to calculate the orientation of
        :return: orientation (rad) of the lanelet
        """
        # TODO: This method could make use of the general relative orientation function or could be fully replaced by
        #  the functions moved to the route planner
        # TODO: This function is a duplicate with the function in EgoObservation

        idx = spatial.KDTree(polyline).query(position)[1]
        if idx < position.shape[0] - 1:
            orientation = np.arccos((polyline[idx + 1, 0] - polyline[idx, 0]) /
                                    np.linalg.norm(polyline[idx + 1] - polyline[idx]))
            sign = np.sign(polyline[idx + 1, 1] - polyline[idx, 1])
        else:
            orientation = np.arccos((polyline[idx, 0] - polyline[idx - 1, 0]) /
                                    np.linalg.norm(polyline[idx] - polyline[idx - 1]))
            sign = np.sign(polyline[idx, 1] - polyline[idx - 1, 1])
        if sign >= 0:
            orientation = np.abs(orientation)
        else:
            orientation = -np.abs(orientation)
        # orientation = shift_orientation(orientation)
        return orientation

    def _create_lanes_for_lanelet_network(self) -> List:
        """
        Searches drivable lanes in the lanelet netwirk of the scenario
        :return: drivable lane list (sub lanelet ids and merged lanelet)
        """
        # TODO: generalize this for other datasets than inD
        lanes = []
        for lanelet in self.scenario.lanelet_network.lanelets:
            if lanelet.predecessor:
                for predecessor in lanelet.predecessor:
                    # find the lanelet without predecessor (iteration ?)
                    if not self.scenario.lanelet_network.find_lanelet_by_id(predecessor).predecessor:
                        (lane_lanelets, sub_lanelets_ids) = Lanelet.all_lanelets_by_merging_successors_from_lanelet(
                            self.scenario.lanelet_network.find_lanelet_by_id(predecessor),
                            self.scenario.lanelet_network, 1000)
                        # in InD dataset, there may be multiple lanelets to go
                        for merged_lane_lanelet, sub_lanelet_ids in zip(lane_lanelets, sub_lanelets_ids):
                            identifer = '_'.join(self.scenario.scenario_id.__str__().split('_')[0:2])
                            # fix imperfect design in the commonroad model
                            # planning problem: DEU_AAH-1_80029_T-1
                            if identifer == 'DEU_AAH-1':
                                self.weight_coefficient = 12
                                if 153 not in sub_lanelet_ids:
                                    if sub_lanelet_ids[0] == 112 and sub_lanelet_ids[-1] == 102:
                                        smooth_factor = 1.7
                                    else:
                                        smooth_factor = 1.5
                                    reference_path = extrapolate_resample_polyline(merged_lane_lanelet.center_vertices)
                                    reference_path = self._smoothing_reference_path(reference_path,
                                                                                    smooth_factor=smooth_factor)
                                    # reference_path = merged_lane_lanelet.center_vertices
                                    # ref_cossy = CurvilinearCoordinateSystem(reference_path)
                                    ref_cossy = CoordinateSystem(reference_path)
                                    lanes.append([sub_lanelet_ids, merged_lane_lanelet, reference_path, ref_cossy])
                            elif identifer == 'DEU_AAH-2':
                                self.weight_coefficient = 8
                                reference_path = extrapolate_resample_polyline(merged_lane_lanelet.center_vertices)
                                reference_path = self._smoothing_reference_path(reference_path, smooth_factor=1.0)
                                # reference_path = merged_lane_lanelet.center_vertices
                                # ref_cossy = CurvilinearCoordinateSystem(reference_path)
                                ref_cossy = CoordinateSystem(reference_path)
                                lanes.append([sub_lanelet_ids, merged_lane_lanelet, reference_path, ref_cossy])
                            elif identifer == 'DEU_AAH-3':
                                self.weight_coefficient = 8
                                # remove unuseful road
                                if sub_lanelet_ids[0] != 149:
                                    reference_path = extrapolate_resample_polyline(merged_lane_lanelet.center_vertices)
                                    if sub_lanelet_ids[0] == 114 and sub_lanelet_ids[-1] in [154, 146]:
                                        reference_path = self._smoothing_reference_path(reference_path,
                                                                                        smooth_factor=0.0)
                                    else:
                                        reference_path = self._smoothing_reference_path(reference_path,
                                                                                        smooth_factor=0.7)
                                    # reference_path = merged_lane_lanelet.center_vertices
                                    # ref_cossy = CurvilinearCoordinateSystem(reference_path)
                                    ref_cossy = CoordinateSystem(reference_path)
                                    lanes.append([sub_lanelet_ids, merged_lane_lanelet, reference_path, ref_cossy])
                            else:
                                reference_path = extrapolate_resample_polyline(merged_lane_lanelet.center_vertices)
                                reference_path = self._smoothing_reference_path(reference_path, smooth_factor=1.0)
                                # reference_path = merged_lane_lanelet.center_vertices
                                # ref_cossy = CurvilinearCoordinateSystem(reference_path)
                                ref_cossy = CoordinateSystem(reference_path)
                                lanes.append([sub_lanelet_ids, merged_lane_lanelet, reference_path, ref_cossy])

            elif not lanelet.predecessor and not lanelet.successor:
                reference_path = extrapolate_resample_polyline(lanelet.center_vertices)
                reference_path = self._smoothing_reference_path(reference_path, smooth_factor=1.2)
                # reference_path = lanelet.center_vertices
                # ref_cossy = CurvilinearCoordinateSystem(reference_path)
                ref_cossy = CoordinateSystem(reference_path)
                lanes.append([[lanelet.lanelet_id], lanelet, reference_path, ref_cossy])

        return lanes

    def _smoothing_reference_path(self, reference_path: np.ndarray, smooth_factor=None) -> np.ndarray:
        """
        Smooths a given reference polyline for lower curvature rates. The smoothing is done using splines from scipy.
        :param reference_path: The reference_path to smooth [array]
        :return: The smoothed reference
        """
        # TODO: check if duplicate with reactive planner utils
        if smooth_factor is None:
            smooth_factor = self.smooth_factor
        # generate a smooth reference path
        transposed_reference_path = reference_path.T
        # how to generate index okay
        okay = np.where(
            np.abs(np.diff(transposed_reference_path[0]))
            + np.abs(np.diff(transposed_reference_path[1]))
            > 0
        )
        xp = np.r_[transposed_reference_path[0][okay], transposed_reference_path[0][-1]]
        yp = np.r_[transposed_reference_path[1][okay], transposed_reference_path[1][-1]]

        curvature = compute_curvature_from_polyline(np.array([xp, yp]).T)

        # set weights for interpolation:
        # see details: https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.splprep.html
        weights = np.exp(-self.weight_coefficient * (abs(curvature) - np.min(abs(curvature))))
        # B spline interpolation
        tck, u = splprep([xp, yp], s=smooth_factor, w=weights)
        u_new = np.linspace(u.min(), u.max(), 2000)
        x_new, y_new = splev(u_new, tck, der=0)

        return np.array([x_new, y_new]).transpose()


class CoordinateSystem():
    # TODO: check if this class is necessary or can be eliminated
    def __init__(self, reference: np.ndarray):
        self._reference = reference
        # self._ccosy = create_coordinate_system_from_polyline(reference)
        self._ccosy = CurvilinearCoordinateSystem(reference)
        self._ref_pos = compute_pathlength_from_polyline(reference)
        self._ref_curv = compute_curvature_from_polyline(reference)
        self._ref_theta = compute_orientation_from_polyline(reference)
        self._ref_curv_d = np.gradient(self._ref_curv, self._ref_pos)

    def reference(self) -> np.ndarray:
        return self._reference

    def ccosy(self) -> CurvilinearCoordinateSystem: #TrapezoidCoordinateSystem:
        return self._ccosy

    def ref_pos(self) -> np.ndarray:
        return self._ref_pos

    def ref_curv(self) -> np.ndarray:
        return self._ref_curv

    def ref_curv_d(self) -> np.ndarray:
        return self._ref_curv_d

    def ref_theta(self) -> np.ndarray:
        return self._ref_theta

    def convert_to_cartesian_coords(self, s: float, d: float) -> np.ndarray:
        try:
            cartesian_coords = self._ccosy.convert_to_cartesian_coords(s, d)
        except:
            cartesian_coords = None

        return cartesian_coords

    def convert_to_curvilinear_coords(self, x: float, y: float) -> np.ndarray:
        return self._ccosy.convert_to_curvilinear_coords(x, y)


def extrapolate_resample_polyline(
    polyline: np.ndarray, step: float = 2.0
) -> np.ndarray:
    """
    Current ccosy (https://gitlab.lrz.de/cps/commonroad-curvilinear-coordinate-system/-/tree/development) creates
    wrong projection domain if polyline has large distance between waypoints --> resampling;
    initial and final points are not within projection domain -> extrapolation
    :param polyline: polyline to be used to create ccosy
    :param step: step distance for resampling
    :return: extrapolated and resampled polyline
    """
    # TODO: check if this method is necessary or can be eliminated by importing others
    p = np.poly1d(np.polyfit(polyline[:2, 0], polyline[:2, 1], 1))

    x = 2 * polyline[0, 0] - polyline[1, 0]
    a = np.array([[x, p(x)]])
    polyline = np.concatenate((a, polyline), axis=0)

    # extrapolate final point
    p = np.poly1d(np.polyfit(polyline[-2:, 0], polyline[-2:, 1], 1))

    # x = 2 * polyline[-1, 0] - polyline[-2, 0]
    # TODO: extend more for learning
    x = 100 * polyline[-1, 0] - 99 * polyline[-2, 0]
    a = np.array([[x, p(x)]])
    polyline = np.concatenate((polyline, a), axis=0)

    return resample_polyline(polyline, step=step)


from collections import OrderedDict
from typing import Union, Dict, List, Tuple, Set, Optional

import commonroad_dc.pycrcc as pycrcc
import gym
import numpy as np
from commonroad.common.util import make_valid_orientation
from commonroad.geometry.shape import Polygon, Rectangle
from commonroad.scenario.lanelet import Lanelet, LaneletType
from commonroad.scenario.trajectory import Trajectory, State, CustomState
from commonroad.scenario.obstacle import Obstacle, SignalState, ObstacleRole, ObstacleType, StaticObstacle
from commonroad.scenario.scenario import Scenario
from commonroad.visualization.mp_renderer import MPRenderer, ZOrders
from commonroad.visualization.util import LineDataUnits
from commonroad.scenario.traffic_sign import TrafficSignIDGermany
from commonroad_rl.tools.trajectory_classification import TrajectoryType, classify_trajectory
from numpy import ndarray
from commonroad_dc.pycrccosy import CurvilinearCoordinateSystem
from shapely.geometry import Point, LineString

from commonroad_rl.gym_commonroad.action.vehicle import Vehicle
from commonroad_rl.gym_commonroad.observation.observation import Observation
from commonroad_rl.gym_commonroad.utils.scenario import approx_orientation_vector
from commonroad_rl.gym_commonroad.utils.navigator import Navigator
from commonroad_rl.gym_commonroad.utils.conflict_zone import ConflictZone


class SurroundingObservation(Observation):

    def __init__(self, configs: Dict, config_name: str = "surrounding_configs"):
        # Read configs
        
        hybrid_reward_configs = configs["reward_configs"]["hybrid_reward"]
        
        configs = configs[config_name]

        self.observe_is_collision: bool = configs.get("observe_is_collision", True)

        self.observe_vehicle_type: bool = configs.get("observe_vehicle_type", False)
        self.num_vehicle_types: int = 7
        self.observe_vehicle_lights: bool = configs.get("observe_vehicle_lights", False)

        self.fast_distance_calculation: bool = configs.get("fast_distance_calculation", True)

        self.observe_lane_rect_surrounding: bool = configs.get("observe_lane_rect_surrounding", False)
        self.lane_rect_sensor_range_length: float = configs.get("lane_rect_sensor_range_length", 100)
        self.lane_rect_sensor_range_width: float = configs.get("lane_rect_sensor_range_width", 7)

        self.observe_lane_circ_surrounding: bool = configs.get("observe_lane_circ_surrounding", False)
        self.lane_circ_sensor_range_radius: float = configs.get("lane_circ_sensor_range_radius", 100)

        self.observe_lidar_circle_surrounding: bool = configs.get("observe_lidar_circle_surrounding", False)
        self.lidar_circle_num_beams: int = configs.get("lidar_circle_num_beams", 20)
        self.lidar_sensor_radius: float = configs.get("lidar_sensor_radius", 100)
        self.observe_lane_change: bool = configs.get("observe_lane_change", False)

        self.observe_relative_priority: bool = configs.get("observe_relative_priority", False)

        self.observe_intersection_velocities = configs.get("observe_intersection_velocities", False)
        self.observe_intersection_distances = configs.get("observe_intersection_distances", False)
        self.observe_ego_distance_intersection = configs.get("observe_ego_distance_intersection", False)
        self.dummy_dist_intersection:float =  configs.get("dummy_dist_intersection",50.)


        self.max_obs_dist: float = 0.0
        if self.observe_lane_circ_surrounding:
            self.max_obs_dist = self.lane_circ_sensor_range_radius
        elif self.observe_lane_rect_surrounding:
            self.max_obs_dist \
                = np.sqrt((self.lane_rect_sensor_range_length / 2) ** 2 + (self.lane_rect_sensor_range_width / 2) ** 2)
        elif self.observe_lidar_circle_surrounding:
            self.max_obs_dist = self.lidar_sensor_radius

        assert sum([self.observe_lidar_circle_surrounding, self.observe_lane_rect_surrounding,
                    self.observe_lane_circ_surrounding]) <= 1, "Only one kind of surrounding observation can be active!"

        self.reward_safe_distance_coef: float = hybrid_reward_configs.get("reward_safe_distance_coef")
        

        self._local_ccosy = None
        self._scenario: Scenario = None
        self._current_time_step = None
        self._ego_state = None
        self._last_ego_lanelet_id = None
        self._surrounding_area = None
        self._detected_obstacle_states = None
        self._surrounding_beams = None
        self._detection_points = None
        self.detected_obstacles = None
        self.lanelet_dict = None
        self.all_lanelets_set = None
        self.observation_dict = OrderedDict()

    def build_observation_space(self) -> OrderedDict:
        observation_space_dict = OrderedDict()

        # Lane-based circular/rectangle surrounding observation
        if self.observe_lane_rect_surrounding or self.observe_lane_circ_surrounding:
            observation_space_dict["lane_based_v_rel"] = gym.spaces.Box(-np.inf, np.inf, (6,), dtype=np.float32)
            observation_space_dict["lane_based_p_rel"] = gym.spaces.Box(-self.max_obs_dist, self.max_obs_dist, (6,),
                                                                        dtype=np.float32)
            if self.observe_vehicle_type:
                observation_space_dict["vehicle_type"] = gym.spaces.Box(0, self.num_vehicle_types - 1, (6,),
                                                                        dtype=np.int8)
            if self.observe_vehicle_lights:
                observation_space_dict["vehicle_signals"] = gym.spaces.Box(-1, 1, (6,), dtype=np.int8)
        # Lidar-based elliptical surrounding observation
        elif self.observe_lidar_circle_surrounding:
            num_beams = self.lidar_circle_num_beams
            observation_space_dict["lidar_circle_dist_rate"] = gym.spaces.Box(-np.inf, np.inf, (num_beams,),
                                                                              dtype=np.float32)
            observation_space_dict["lidar_circle_dist"] = gym.spaces.Box(-self.max_obs_dist, self.max_obs_dist,
                                                                         (num_beams,), dtype=np.float32)

            if self.observe_vehicle_type:
                observation_space_dict["vehicle_type"] = gym.spaces.Box(0, self.num_vehicle_types - 1, (num_beams,),
                                                                        dtype=np.int8)
            if self.observe_vehicle_lights:
                observation_space_dict["vehicle_signals"] = gym.spaces.Box(-1, 1, (num_beams,), dtype=np.int8)

            if self.reward_safe_distance_coef != 0:
                observation_space_dict["dist_lead_follow_rel"] = gym.spaces.Box(-self.max_obs_dist, self.max_obs_dist,
                                                                                (2,),
                                                                                dtype=np.float32)
            if self.observe_relative_priority:
                observation_space_dict["rel_prio_lidar"] = gym.spaces.Box(-1, 1, (num_beams,), dtype=np.float32)
        if self.observe_is_collision:
            observation_space_dict["is_collision"] = gym.spaces.Box(0, 1, (1,), dtype=np.float32)

        if self.observe_intersection_velocities:
            observation_space_dict["intersection_velocities"] = gym.spaces.Box(-np.inf, np.inf, (6,), dtype=np.float32)

        if self.observe_intersection_distances:
            observation_space_dict["intersection_distances"] = gym.spaces.Box(-self.dummy_dist_intersection,
                                                                              self.dummy_dist_intersection, (6,),
                                                                              dtype=np.float32)
        if self.observe_ego_distance_intersection:
            observation_space_dict["ego_distance_intersection"] = gym.spaces.Box(-np.inf, np.inf, (2,),
                                                                                 dtype=np.float32)

        if self.observe_lane_change:
            observation_space_dict["lane_change"] = gym.spaces.Box(0, 1, (1,), dtype=np.float32)
        return observation_space_dict

    def observe(self, scenario: Scenario, ego_vehicle: Vehicle, time_step: int,
                connected_lanelet_dict: Union[None, Dict] = None, ego_lanelet: Union[Lanelet, None] = None,
                collision_checker: Union[None, pycrcc.CollisionChecker] = None,
                local_ccosy: Union[None, CurvilinearCoordinateSystem] = None, 
                conflict_zone: ConflictZone = None,
                ego_lanelet_ids: List[int] = []) -> Union[ndarray, Dict]:

        # update scenario and time_step
        self._scenario = scenario
        self._current_time_step = time_step
        self._ego_state = ego_vehicle.state
        if not self.fast_distance_calculation:
            self._ego_shape = Rectangle(length=ego_vehicle.parameters.l,
                                        width=ego_vehicle.parameters.w,
                                        center=self._ego_state.position,
                                        orientation=self._ego_state.orientation).shapely_object
        self._collision_checker = collision_checker
        self._local_ccosy = local_ccosy
        self.lanelet_dict, self.all_lanelets_set = SurroundingObservation.get_nearby_lanelet_id(connected_lanelet_dict,
                                                                                                ego_lanelet)
        ego_vehicle_lat_position = None

        if self.observe_lane_rect_surrounding or self.observe_lane_circ_surrounding:
            # construct sensing area
            if self.observe_lane_rect_surrounding:
                self._surrounding_area = pycrcc.RectOBB(self.lane_rect_sensor_range_length / 2,
                                                        self.lane_rect_sensor_range_width / 2,
                                                        self._ego_state.orientation if hasattr(
                                                            self._ego_state,
                                                            "orientation") else np.arctan2(
                                                            self._ego_state.velocity_y,
                                                            self._ego_state.velocity),
                                                        self._ego_state.position[0],
                                                        self._ego_state.position[1])
            else:
                self._surrounding_area = pycrcc.Circle(self.lane_circ_sensor_range_radius,
                                                       self._ego_state.position[0],
                                                       self._ego_state.position[1])

            ego_vehicle_lat_position, self._detected_obstacle_states, self.detected_obstacles = \
                self._get_surrounding_obstacles_lane_based(self._surrounding_area)

        elif self.observe_lidar_circle_surrounding:
            self._surrounding_area = pycrcc.Circle(self.lidar_sensor_radius, self._ego_state.position[0],
                                                   self._ego_state.position[1])
            self.detected_obstacles = self._get_surrounding_obstacles_lidar_circle()
            if self.reward_safe_distance_coef != 0:
                self._add_leading_following_distance_lidar_lane()
            if self.observe_relative_priority:
                self._add_relative_priority(self.detected_obstacles, ego_lanelet, ego_vehicle)

        if self.observe_lane_change:
            self._detect_lane_change(ego_lanelet_ids)

        if self.observe_is_collision:
            is_collision = self._check_collision(collision_checker, ego_vehicle)
            self.observation_dict["is_collision"] = np.array([is_collision])

        if self.observe_intersection_velocities or self.observe_intersection_distances:
            if self.observe_lane_circ_surrounding:
                intersection_observation = conflict_zone.generate_intersection_observation(self.conflict_obstacles_information)
                if self.observe_intersection_velocities:
                    self._get_intersection_velocities(intersection_observation)

                if self.observe_intersection_distances:
                    self._get_intersection_distances(intersection_observation)
            else:
                #TODO: Remove this constraint
                print('ERROR:SurroundingObservation: Intersection observations currently only selectable together '
                      'with circle lane based observation')
                raise ValueError

        if self.observe_ego_distance_intersection:
            s_near, s_far = conflict_zone.get_ego_intersection_observation(self._ego_state.position)
            self.observation_dict["ego_distance_intersection"] = np.array([s_near, s_far])

        return self.observation_dict, ego_vehicle_lat_position

    def draw(self, render_configs: Dict, render: MPRenderer, terminated: bool = False):
        # Mark surrounding obstacles (Only if corresponding observations are available)
        # Lane-based surrounding rendering
        # surrounding areas
        if render_configs["render_surrounding_area"] and (
                self.observe_lane_rect_surrounding or self.observe_lane_circ_surrounding):
            self._surrounding_area.draw(render, draw_params={"facecolor": "lightblue",
                                                             "edgecolor": "lightblue", "opacity": 0.5})

        # detected obstacles
        if render_configs["render_surrounding_obstacles_lane_based"] and (
                self.observe_lane_rect_surrounding or self.observe_lane_circ_surrounding):
            colors = ["r", "y", "k", "r", "y", "k"]
            # o_left_follow, o_same_follow, o_right_follow, o_left_lead, o_same_lead, o_right_lead
            for obs, color in zip(self._detected_obstacle_states, colors):
                if obs is not None:
                    render.dynamic_artists.append(
                        LineDataUnits([obs.position[0]], [obs.position[1]], color=color, marker="*",
                                      zorder=ZOrders.OBSTACLES+1, label="surrounding_obstacles_lane_based"))

        # Lidar-based elliptical surrounding rendering
        if self.observe_lidar_circle_surrounding and render_configs["render_lidar_circle_surrounding_beams"]:
            for (beam_start, beam_length, beam_angle) in self._surrounding_beams:
                center = beam_start + 0.5 * beam_length * approx_orientation_vector(beam_angle)
                beam_angle = make_valid_orientation(beam_angle)
                beam_draw_object = Rectangle(length=beam_length, width=0.1, center=center, orientation=beam_angle)
                beam_draw_object.draw(render)

        if self.observe_lidar_circle_surrounding and render_configs["render_lidar_circle_surrounding_obstacles"]:
            for idx, detection_point in enumerate(self._detection_points):
                    render.dynamic_artists.append(
                    LineDataUnits(detection_point[0], detection_point[1], color="b", marker="1",
                                  zorder=ZOrders.INDICATOR_ADD, label="surrounding_obstacles_lidar_based"))

        # TODO: add rendering for intersection observations
        # use conflict_region.draw_conflict_region to highlight region
        # also highlight detected intersection vehicles


    def _get_intersection_velocities(self,intersection_observations: List[State]):
        """
        calculates relative velocity observation for intersection vehicles
        :param intersection_observations: list of intersection obstacle states
        return: relative velocity observations for intersection vehicles
        """
        rel_v = np.zeros((6,))
        for k in range(len(intersection_observations)):
            if k < 6:
                rel_v[k] = intersection_observations[k][1]

        self.observation_dict["intersection_velocities"] = rel_v

    def _get_intersection_distances(self, intersection_observations: List[State]):
        """
        calculates relative distance observation for intersection vehicles
        :param intersection_observations: list of intersection obstacle states
        return: relative distance observations for intersection vehicles
        """
        rel_p = self.dummy_dist_intersection * np.ones((6,))
        for k in range(len(intersection_observations)):
            if k < 6:
                rel_p[k] = min(self.dummy_dist_intersection, intersection_observations[k][0])

        self.observation_dict["intersection_distances"] = rel_p


    def _get_surrounding_obstacles_lane_based(self, surrounding_area: Union[pycrcc.RectOBB, pycrcc.Circle]) \
            -> Tuple[np.array, List[State], List[Obstacle]]:
       
        lanelet_ids, obstacle_states, obstacles = self._get_obstacles_in_surrounding_area(surrounding_area)
        obstacle_lanelet, adj_obstacle_states, adj_obstacles = \
            self._filter_obstacles_in_adj_lanelet(lanelet_ids, obstacle_states, obstacles, self.all_lanelets_set)
        rel_vel, rel_pos, detected_states, detected_obstacles, ego_vehicle_lat_position = \
            self._get_rel_v_p_lane_based(obstacle_lanelet, adj_obstacle_states, self.lanelet_dict, adj_obstacles)

        self.observation_dict["lane_based_v_rel"] = np.array(rel_vel)
        self.observation_dict["lane_based_p_rel"] = np.array(rel_pos)

        if self.observe_vehicle_type:
            self._get_vehicle_types(detected_obstacles)
        if self.observe_vehicle_lights:
            self._get_vehicle_lights(detected_obstacles)


        # add self.conflict_obstacles_information
        detected_states_exclude = [state for state in detected_states if state is not None] # TODO: remove after released cr-io
        self.conflict_obstacles_information = [state for state in obstacle_states if state not in detected_states_exclude]
        # if want to include the 6 vehilces that are already detected: comment the above line and uncomment the following line
        # self.conflict_obstacles_information = [state for state in obstacle_states]

        return ego_vehicle_lat_position, detected_states, detected_obstacles

    def _detect_lane_change(self, ego_lanelet_ids: List[int]) -> None:
        self.observation_dict["lane_change"] = np.array([0.0])
        for lanelet_id in ego_lanelet_ids:
            if lanelet_id not in self.lanelet_dict["ego_all"]:
                self.observation_dict["lane_change"] = np.array([1.0])
                return

    def _add_leading_following_distance_lidar_lane(self) -> None:
        """
        Adds the leading and following obstacle distance observation to dist_lead_follow_rel
        """
        lanelet_ids, obstacle_states, obstacles = self._get_obstacles_in_surrounding_area(self._surrounding_area)
        obstacle_lanelet, adj_obstacle_states, adj_obstacles = \
            self._filter_obstacles_in_adj_lanelet(lanelet_ids, obstacle_states, obstacles, self.all_lanelets_set)
        _, rel_pos, _, _, _ = \
            self._get_rel_v_p_lane_based(obstacle_lanelet, adj_obstacle_states, self.lanelet_dict, adj_obstacles)
        self.observation_dict["dist_lead_follow_rel"] = np.array([rel_pos[4], rel_pos[1]])

    def _add_relative_priority(self, obstacles: List[Obstacle], ego_lanelet: Lanelet,
                               ego_vehicle: Vehicle, turn_threshold=0.02, scan_range=20.):
        """
        Adds relative priority to the observation dict. {-1, 0, 1} for {yield, same, priority} respectively
        :param obstacles: a list of detected obstacles
        """
        # Classify ego trajectory from the reference path (project position onto reference path)
        if self._local_ccosy.cartesian_point_inside_projection_domain(
                self._ego_state.position[0], self._ego_state.position[1]):
            ego_curv_coords = self._local_ccosy.convert_to_curvilinear_coords(self._ego_state.position[0],
                                                                              self._ego_state.position[1])
            ego_proj_pos = self._local_ccosy.convert_to_cartesian_coords(ego_curv_coords[0], 0)
            t_type = self.trajectory_type_from_path(
                self._local_ccosy.reference_path(), turn_threshold, pycrcc.Circle(scan_range, ego_proj_pos[0],
                                                                                  ego_proj_pos[1]))
        # TODO: possible improvement - find the nearest possible point in the domain or use the last state
        #  in the domain as reference
        # If the point is not in the projection domain, classify the entire reference path
        else:
            t_type = self.trajectory_type_from_path(self._local_ccosy.reference_path(), turn_threshold)
        # Calculate ego priority according to its trajectory class and signs
        ego_lanelet_priority = self._detect_lanelet_priority(ego_lanelet, t_type)
        obstacle_rel_lanelet_priorities = []
        # Iterate over the obstacles
        for obs in obstacles:
            if obs is None or obs.obstacle_type == ObstacleRole.STATIC:
                obstacle_rel_lanelet_priorities.append(1.0)
                continue
            obstacle_state = obs.state_at_time(self._current_time_step)
            obs_occupied_lanelet_id = list(obs.prediction.shape_lanelet_assignment[self._current_time_step])
            if len(obs_occupied_lanelet_id) > 1:
                obs_occupied_lanelet_id = Navigator.sorted_lanelet_ids(
                    obs_occupied_lanelet_id, obstacle_state.orientation, obstacle_state.position, self._scenario)[0]
            elif len(obs_occupied_lanelet_id) == 1:
                obs_occupied_lanelet_id = obs_occupied_lanelet_id[0]
            elif len(obs_occupied_lanelet_id) == 0:
                obstacle_rel_lanelet_priorities.append(0.0)
                continue

            obs_lanelet = self._scenario.lanelet_network.find_lanelet_by_id(obs_occupied_lanelet_id)

            obstacle_turning_signal = obs.signal_state_at_time_step(self._current_time_step)

            if obstacle_turning_signal is None:
                obs_t_type = TrajectoryType.STRAIGHT
            elif obstacle_turning_signal.indicator_left:
                obs_t_type = TrajectoryType.LEFT
            elif obstacle_turning_signal.indicator_right:
                obs_t_type = TrajectoryType.RIGHT
            # else:
            #     obs_t_type = TrajectoryType.STRAIGHT

            # Classify the obstacle priority based on its trajectory class and signs
            obstacle_lanelet_priority = self._detect_lanelet_priority(obs_lanelet, obs_t_type)
            intersection_dict = self._scenario.lanelet_network.map_inc_lanelets_to_intersections
            # Check if ego and the obstacle are approaching the same intersection
            # or if the obstacle is already in an intersection
            if ego_lanelet.lanelet_id in intersection_dict \
                    and ((obs_lanelet.lanelet_id in intersection_dict and
                          intersection_dict[ego_lanelet.lanelet_id] == intersection_dict[obs_lanelet.lanelet_id])
                         or obs_lanelet.lanelet_type == LaneletType.INTERSECTION):
                # If an obstacle is in the middle of an intersection -> yield
                if obs_lanelet.lanelet_type == LaneletType.INTERSECTION:
                    obstacle_rel_lanelet_priorities.append(-1.0)
                # Obstacle in the same lanelet as ego -> same priority
                # TODO: include adjacent lanelets -> obstacles can be ignored as well
                elif obs_lanelet.lanelet_id == ego_lanelet.lanelet_id:
                    obstacle_rel_lanelet_priorities.append(0.0)
                # The signs match -> check right before left
                elif ego_lanelet_priority in [4, 5, 6] and obstacle_lanelet_priority in [4, 5, 6] or \
                        self._matching_signs(ego_lanelet, obs_lanelet):
                    # Right before left rule applies, check by orientation
                    obs_state = obs.state_at_time(self._current_time_step)
                    ego_state = ego_vehicle.state
                    ego_state_orientation = ego_state.orientation if hasattr(ego_state, "orientation") else \
                        np.arctan2(ego_state.velocity_y, ego_state.velocity)
                    rel_orientation = make_valid_orientation(obs_state.orientation - ego_state_orientation)
                    if np.isclose(rel_orientation, np.pi / 2, atol=np.pi * 1 / 6):
                        obstacle_rel_lanelet_priorities.append(-1.0)
                        continue
                    elif np.isclose(rel_orientation, 1.5 * np.pi, atol=np.pi * 1 / 6):
                        obstacle_rel_lanelet_priorities.append(1.0)
                        continue
                    # obstacle_rel_lanelet_priorities.append(-1.0)
                # Check sign + driving direction rules
                if obstacle_lanelet_priority < ego_lanelet_priority:
                    obstacle_rel_lanelet_priorities.append(1.0)
                elif obstacle_lanelet_priority > ego_lanelet_priority:
                    obstacle_rel_lanelet_priorities.append(-1.0)
                else:
                    obstacle_rel_lanelet_priorities.append(0.0)
            else:
                obstacle_rel_lanelet_priorities.append(1.0)
        self.observation_dict["rel_prio_lidar"] = np.array(obstacle_rel_lanelet_priorities)

    @staticmethod
    def trajectory_type_from_path(ref_path: ndarray, turn_threshold=0.02, traj_area=None) -> TrajectoryType:
        """
        Classifies the trajectory of a path (2d polyline) at the turn threshold. Filters points that aren't in
        the surrounding area
        :param ref_path: 2d polyline path that gets classified
        :param turn_threshold: the minimum curvature of the path to be classified as a turn
        :param traj_area: shapely object of the area in which the trajectory's analyzed
        """
        ref_path_before = ref_path
        if traj_area:
            ref_path_points = []
            for point in ref_path:
                obstacle_point = pycrcc.Point(point[0], point[1])
                if traj_area.collide(obstacle_point):
                    ref_path_points.append(point)
            ref_path = ref_path_points

        state_list = []
        dummy_velocity = 1.0
        time_step = 0
        for point in ref_path:
            state_list.append(CustomState(position=point, velocity=dummy_velocity, time_step=time_step))
            time_step += 1
        assert len(state_list) != 0, "ref_path is " + str(ref_path) + " vs before " + str(
            ref_path_before) + " shape: (r,x,y)" + str(traj_area.r()) + " " + str(traj_area.x()) + " " + str(
            traj_area.y())
        traj = Trajectory(0, state_list)
        t_type = classify_trajectory(traj, min_velocity=dummy_velocity - 1.0, turn_threshold=turn_threshold)
        return t_type

    def _matching_signs(self, ego_lanelet: Lanelet, obs_lanelet: Lanelet):
        sign_ids_obs = list(self._scenario.lanelet_network.find_traffic_sign_by_id(s).
                            traffic_sign_elements[0].traffic_sign_element_id for s in obs_lanelet.traffic_signs)
        sign_ids_ego = list(self._scenario.lanelet_network.find_traffic_sign_by_id(s).
                            traffic_sign_elements[0].traffic_sign_element_id for s in ego_lanelet.traffic_signs)
        return any(s_id == s_id_ego and s_id in [TrafficSignIDGermany.YIELD, TrafficSignIDGermany.STOP,
                                                 TrafficSignIDGermany.RIGHT_OF_WAY, TrafficSignIDGermany.PRIORITY]
                   for s_id in sign_ids_obs for s_id_ego in sign_ids_ego)

    def _detect_lanelet_priority(self, lanelet: Lanelet, traj_type: TrajectoryType) -> int:
        """
        Returns a priority number between 1 and 9. The larger the number, the earlier the vehicle is allowed to drive
        Only includes lanelet priority and trajectory (no right before left rule)
        """
        sign_ids = list(self._scenario.lanelet_network.find_traffic_sign_by_id(s).
                        traffic_sign_elements[0].traffic_sign_element_id for s in lanelet.traffic_signs)

        if any(s_id in [TrafficSignIDGermany.YIELD, TrafficSignIDGermany.STOP] for s_id in sign_ids):
            if traj_type == TrajectoryType.LEFT:
                return 1
            elif traj_type == TrajectoryType.RIGHT:
                return 2
            else:
                return 3
        elif any(s_id in [TrafficSignIDGermany.RIGHT_OF_WAY, TrafficSignIDGermany.PRIORITY] for s_id in sign_ids):
            if traj_type == TrajectoryType.LEFT:
                return 7
            elif traj_type == TrajectoryType.RIGHT:
                return 8
            else:
                return 9
        else:
            if traj_type == TrajectoryType.LEFT:
                return 4
            elif traj_type == TrajectoryType.RIGHT:
                return 5
            else:
                return 6

    def _get_surrounding_obstacles_lidar_circle(self) -> List[Obstacle]:

        obstacle_shapes, detected_obstacles = self._get_obstacle_shapes_in_surrounding_area(self._surrounding_area)

        # Create beam shapes (shapely line strings) around the ego vehicle, forming an ellipse sensing area as a whole
        surrounding_beams_ego_vehicle = []
        beam_start = self._ego_state.position
        for i in range(self.lidar_circle_num_beams):
            theta = i * (2 * np.pi / self.lidar_circle_num_beams)
            x_delta = self.lidar_sensor_radius * np.cos(theta)
            y_delta = self.lidar_sensor_radius * np.sin(theta)
            beam_length = np.sqrt(x_delta ** 2 + y_delta ** 2)
            beam_angle = self._ego_state.orientation + theta
            surrounding_beams_ego_vehicle.append((beam_start, beam_length, beam_angle))

        obstacle_distances, observed_obstacles \
            = self._get_obstacles_with_surrounding_beams(obstacle_shapes, detected_obstacles,
                                                         surrounding_beams_ego_vehicle)

        distances, distance_rates, detection_points = self._get_distances_lidar_based(surrounding_beams_ego_vehicle,
                                                                                      obstacle_distances)

        self.observation_dict["lidar_circle_dist_rate"] = np.array(distance_rates)
        self.observation_dict["lidar_circle_dist"] = np.array(distances)

        if self.observe_vehicle_type:
            self._get_vehicle_types(observed_obstacles)
        if self.observe_vehicle_lights:
            self._get_vehicle_lights(observed_obstacles)
        self._surrounding_beams = surrounding_beams_ego_vehicle
        self._detection_points = detection_points

        return observed_obstacles

    @staticmethod
    def _check_collision(collision_checker: pycrcc.CollisionChecker, ego_vehicle: Vehicle) -> bool:
        collision_ego_vehicle = ego_vehicle.collision_object
        return collision_checker.collide(collision_ego_vehicle)

    def _get_obstacle_shapes_in_surrounding_area(self, surrounding_area: pycrcc.Shape) \
            -> Tuple[List[Polygon], List[Obstacle]]:
        """
        Get the occupancy shape and states and lanelet ids of all obstacles
        within the range of surrounding area of ego vehicle.
        :param surrounding_area: Shapes of pycrcc classes
        :return: List of obstacle shapely shapes
        """
        obstacle_shapes = []
        detected_obstacles = []
        dyn_obstacles, static_obstacles = self._scenario.dynamic_obstacles, self._scenario.static_obstacles

        for o in dyn_obstacles:
            if o.initial_state.time_step <= self._current_time_step <= o.prediction.trajectory.final_state.time_step:
                obstacle_state = o.state_at_time(self._current_time_step)
                obstacle_point = pycrcc.Point(obstacle_state.position[0], obstacle_state.position[1])
                if surrounding_area.collide(obstacle_point):
                    obstacle_shapes.append(o.occupancy_at_time(self._current_time_step).shape.shapely_object)
                    detected_obstacles.append(o)

        for o in static_obstacles:
            obstacle_state = o.initial_state
            obstacle_point = pycrcc.Point(obstacle_state.position[0], obstacle_state.position[1])
            if surrounding_area.collide(obstacle_point):
                obstacle_shapes.append(o.occupancy_at_time(self._current_time_step).shape.shapely_object)
                detected_obstacles.append(o)

        return obstacle_shapes, detected_obstacles

    def _get_obstacles_with_surrounding_beams(self, obstacle_shapes: List[Polygon], obstacles: List[Obstacle],
                                              surrounding_beams: List[Tuple[float, float, float]]) \
            -> Tuple[np.ndarray, List[Optional[Obstacle]]]:
        """
        Get the distance to the nearest obstacles colliding with LIDAR beams

        :param obstacle_shapes: Obstacle shapes that detected with given sensing area
        :param obstacles: The obstacles belonging to the obstacle_shapes
        :param surrounding_beams: List of beams as start point, length and angle
        :return: List of obstacle states
        """
        obstacle_distances = np.zeros(len(surrounding_beams))
        observed_obstacles = [None] * len(surrounding_beams)
        # For each beam, record all collisions with obstacles first, and report the one being closest to the ego vehicle
        ego_vehicle_center_shape = Point(self._ego_state.position)
        for (i, (beam_start, beam_length, beam_angle)) in enumerate(surrounding_beams):
            beam_vec = approx_orientation_vector(beam_angle) * beam_length
            # asLineString recycles C-array as explained
            # beam = asLineString(np.array([beam_start, beam_start + beam_vec]))
            beam = LineString([beam_start, beam_start + beam_vec])
            obstacle_candidate_dist = self.max_obs_dist
            obstacle_candidate = None

            for j, obstacle_shape in enumerate(obstacle_shapes):
                # TODO also support Shapegroups without shapely_object attribute
                if beam.intersects(obstacle_shape):
                    dist = ego_vehicle_center_shape.distance(beam.intersection(obstacle_shape))
                    if dist < obstacle_candidate_dist:
                        obstacle_candidate_dist = dist
                        obstacle_candidate = obstacles[j]

            obstacle_distances[i] = obstacle_candidate_dist
            observed_obstacles[i] = obstacle_candidate

        return obstacle_distances, observed_obstacles

    def _get_distances_lidar_based(self, beams: List[Tuple[np.ndarray, float, float]],
                                   obstacle_distances: Union[List[float], np.ndarray]) \
            -> Tuple[np.ndarray, np.ndarray, List[np.ndarray]]:
        """
        calculate the points in the obstacle where the lidar beam hit
        calculate the length/distance of the ego_state to that point
        and calculate the change of the distances to the obstacles over the last time_step

        :param beams: List of beams as start coordinates, length and angle
        :param obstacle_distances: distance of the beam to the detected point
            if it is equal to self.lidar_sensor_radius no object was detected
        :return:
            dists: same as input obstacle_distances
            dist_rates: change of the distances to the detected obstacle normalized by the time step
            detection_points: point where the beam hit the obstacle

        Examples::
            beams = [(np.array([0,0]),10,0)]
            obstacle_distances = [3]
            dists, dist_rates, detection_points = surrounding_observation._get_distances_lidar_based(beams,
            obstacle_distances)
        """
        # Relative positions
        dists = np.array(obstacle_distances)
        prev_distances = self.observation_dict.get("lidar_circle_dist", np.full(len(beams), self.max_obs_dist))

        # Change rates
        if self._current_time_step == 0:
            dist_rates = np.full(len(beams), 0.0)
        else:
            dist_rates = (prev_distances - dists) / self._scenario.dt

        # detection points
        detection_points = [beam_start + approx_orientation_vector(beam_angle) * dists[i] for
                            i, ((beam_start, _, beam_angle), closest_collision) in
                            enumerate(zip(beams, obstacle_distances))]

        return dists, dist_rates, detection_points

    def _get_obstacles_in_surrounding_area(self, surrounding_area: pycrcc.Shape) \
            -> Tuple[List[int], List[State], List[Obstacle]]:
        """
        Get the states and lanelet ids of all obstacles within the range of surrounding area of ego vehicle.

        :return: List of lanelet ids of obstacles, list of states obstacles
        """
        lanelet_ids, obstacle_states, obstacles = [], [], []
        dyn_obstacles, static_obstacles = self._scenario.dynamic_obstacles, self._scenario.static_obstacles

        # iterate over all dynamic obstacles
        for o in dyn_obstacles:
            # TODO: initial lanelet_assignment missed?
            if o.prediction is not None:
                center_lanelet_ids = list(o.prediction.center_lanelet_assignment.values())
                if o.initial_state.time_step <= self._current_time_step <= o.prediction.trajectory.final_state.time_step:
                    obstacle_state = o.state_at_time(self._current_time_step)
                    obstacle_point = pycrcc.Point(obstacle_state.position[0], obstacle_state.position[1])
                    if surrounding_area.collide(obstacle_point):
                        o_center_lanelet_ids = list(o.prediction.center_lanelet_assignment[self._current_time_step])
                        o_shape_lanelet_ids = list(o.prediction.shape_lanelet_assignment[self._current_time_step])
                        o_lanelet_id = self._get_occupied_lanelet_id(self._scenario, o_center_lanelet_ids, obstacle_state)
                        if o_lanelet_id is None:
                            # use shape lanelet assignment instead
                            o_lanelet_id = self._get_occupied_lanelet_id(self._scenario, o_shape_lanelet_ids,
                                                                         obstacle_state)
                            if o_lanelet_id is None:
                                # neither center or shape locate inside a lanelet
                                # TODO: skip or take last available time step ??
                                non_empty_id_sets = [id_set for id_set in center_lanelet_ids if id_set]
                                if len(non_empty_id_sets) == 0:
                                    continue
                                o_lanelet_id = self._get_occupied_lanelet_id(self._scenario, list(non_empty_id_sets[-1]),
                                                                             obstacle_state)
                        lanelet_ids.append(o_lanelet_id)
                        obstacle_states.append(obstacle_state)
                        obstacles.append(o)
            else: # obstacle only has initial state (interactive scenarios)
                if o.initial_state.time_step == self._current_time_step:
                    obstacle_state = o.initial_state
                    obstacle_point = pycrcc.Point(obstacle_state.position[0], obstacle_state.position[1])
                    if surrounding_area.collide(obstacle_point):
                        o_center_lanelet_ids = list(o.initial_center_lanelet_ids)
                        o_shape_lanelet_ids = list(o.initial_shape_lanelet_ids)
                        # sort lanelet ids by orientation and relevance
                        o_lanelet_id = self._get_occupied_lanelet_id(self._scenario, o_center_lanelet_ids, obstacle_state)
                        if o_lanelet_id is None:
                            # use shape lanelet assignment instead
                            o_lanelet_id = self._get_occupied_lanelet_id(self._scenario, o_shape_lanelet_ids,
                                                                         obstacle_state)
                            if o_lanelet_id is None:
                                # neither center or shape locate inside a lanelet
                                continue
                        lanelet_ids.append(o_lanelet_id)
                        obstacle_states.append(obstacle_state)
                        obstacles.append(o)

        # iterate over all static obstacles
        for o in static_obstacles:
            obstacle_state = o.initial_state
            obstacle_point = pycrcc.Point(obstacle_state.position[0], obstacle_state.position[1])
            if surrounding_area.collide(obstacle_point):
                obstacle_lanelet_ids = list(o.initial_center_lanelet_ids)
                lanelet_id = self._get_occupied_lanelet_id(self._scenario, obstacle_lanelet_ids, obstacle_state)
                lanelet_ids.append(lanelet_id)
                obstacle_states.append(obstacle_state)
                obstacles.append(o)

        return lanelet_ids, obstacle_states, obstacles

    def _get_vehicle_types(self, obstacles: List[Optional[Obstacle]]) -> None:
        """
        Sets the obstacle types in observation_dict for all observed obstacles
        {0,1,2,3,4,5,6} = {Other, Car, Bicycle, Pedestrian, Truck, Bus, Static}

        Note: If the mapping of types to numbers is changed, self.num_vehicle_types should be adjusted accordingly

        :param obstacles: The observed obstacles
        """
        types = [0] * len(obstacles)

        for i, obstacle in enumerate(obstacles):
            if obstacle is None:
                continue

            if obstacle.obstacle_role == ObstacleRole.STATIC:
                types[i] = 6
            elif obstacle.obstacle_type == ObstacleType.CAR:
                types[i] = 1
            elif obstacle.obstacle_type == ObstacleType.BICYCLE:
                types[i] = 2
            elif obstacle.obstacle_type == ObstacleType.PEDESTRIAN:
                types[i] = 3
            elif obstacle.obstacle_type == ObstacleType.TRUCK:
                types[i] = 4
            elif obstacle.obstacle_type == ObstacleType.BUS:
                types[i] = 5

        self.observation_dict["vehicle_type"] = np.array(types)

    def _get_vehicle_lights(self, obstacles: List[Optional[Obstacle]]) -> None:
        """
Im        Sets the turning-lights in observation_dict for all observed obstacles
        {-1, 0, 1} = {Left signal, Off, Right signal}

        :param obstacles: The observed obstacles
        """
        signals = [0] * len(obstacles)

        for i, obstacle in enumerate(obstacles):
            if obstacle is None:
                continue

            signal_state: SignalState = obstacle.signal_state_at_time_step(self._current_time_step)
            if signal_state is None:
                continue
            if signal_state.indicator_right:
                signals[i] = 1
            elif signal_state.indicator_left:
                signals[i] = -1

        self.observation_dict["vehicle_signals"] = np.array(signals)

    @staticmethod
    def _get_occupied_lanelet_id(scenario: Scenario, obstacle_lanelet_ids: List[int], obstacle_state: State) \
            -> Union[None, int]:
        """
        gets most relevant lanelet id from obstacle_lanelet_ids for an obstacle that occupies multiple lanelets

        :param scenario: current scenario
        :param obstacle_lanelet_ids: lanelet ids of lanelets occupied by the obstacle
        :param obstacle_state: current state of the obstacle
        """
        if len(obstacle_lanelet_ids) > 1:
            # select the most relevant lanelet
            return Navigator.sorted_lanelet_ids(
                obstacle_lanelet_ids, obstacle_state.orientation, obstacle_state.position, scenario)[0]
        elif len(obstacle_lanelet_ids) == 1:
            return obstacle_lanelet_ids[0]
        else:
            return None

    @staticmethod
    def _filter_obstacles_in_adj_lanelet(lanelet_ids: List[int], states: List[State], obstacles: List[Obstacle],
                                         all_lanelets_set: Set[int]) -> Tuple[List[int], List[State], List[Obstacle]]:
        """
        filters out obstacles states and their corresponding lanelet id
        where the lanelet id is not in the all_lanelets_set

        :param lanelet_ids: List of lanelet ids of obstacles
        :param states: List of states of obstacles
        :param obstacles: List of obstacles
        :param all_lanelets_set: The set of all lanelet ids in the scenario
        :return: The list of lanelets of obstacles, the list of states
        """
        adj_obstacle_states, obstacle_lanelet, adj_obstacles = [], [], []
        for lanelet_id, state, obstacle in zip(lanelet_ids, states, obstacles):
            if lanelet_id in all_lanelets_set:  # Check if the obstacle is in adj lanelets
                obstacle_lanelet.append(lanelet_id)
                adj_obstacle_states.append(state)
                adj_obstacles.append(obstacle)

        return obstacle_lanelet, adj_obstacle_states, adj_obstacles

    @staticmethod
    def get_rel_v_p_follow_leading(distance_sign: int, distance_abs: float, p_rel_follow: float, p_rel_lead: float,
                                   v_rel_follow: float, v_rel_lead: float, obs_state: State, obstacle: Obstacle,
                                   ego_state: State, o_follow: State, o_lead: State, obstacle_follow: Obstacle,
                                   obstacle_lead: Obstacle) -> \
            Tuple[float, float, State, Obstacle, float, float, State, Obstacle]:
        """
        #TODO maybe change signature to only have a single variable for each leading and following
            e.g. instead of o_follow, o_lead just o
        calculates the relative velocity of leading and following obstacles to the ego vehicle

        :param distance_sign: 1 -> follow, !=1 -> lead
        :param distance_abs: absolut distance of ego vehicle to obstacle
        :param p_rel_follow: max distance of a obstacle that is following the ego vehicle
        :param p_rel_lead: max distance of a obstacle that is leading the ego vehicle
        :param v_rel_follow: relative velocity of the following obstacle
        :param v_rel_lead: relative velocity of the leading obstacle
        :param obs_state: state of the obstacle
        :param ego_state: state of the ego vehicle
        :param o_follow: if the obs_state is following it is stored here
        :param o_lead: if the obs_state is leading it is stored here

        :return
            v_rel_follow: relative velocity to following obstacle
            p_rel_follow: relative position to following obstacle
            o_follow: state of following obstacle
            v_rel_lead: relative velocity to leading obstacle
            p_rel_lead: relative position to leading obstacle
            o_lead: state of leading obstacle
        """
        if isinstance(obstacle, StaticObstacle):
            obs_state.velocity = 0.
        if distance_sign == 1 and distance_abs < p_rel_follow:
            # following vehicle, distance is smaller
            ego_state_orientation = ego_state.orientation if hasattr(ego_state, "orientation") else np.arctan2(
                ego_state.velocity_y, ego_state.velocity)
            delta_orientation = obs_state.orientation - ego_state_orientation
            v_rel_follow = ego_state.velocity - obs_state.velocity * np.cos(delta_orientation)
            p_rel_follow = distance_abs
            o_follow = obs_state
            obstacle_follow = obstacle
        elif distance_sign != 1 and distance_abs < p_rel_lead:
            # leading vehicle, distance is smaller
            ego_state_orientation = ego_state.orientation if hasattr(ego_state, "orientation") else np.arctan2(
                ego_state.velocity_y, ego_state.velocity)
            delta_orientation = obs_state.orientation - ego_state_orientation
            v_rel_lead = obs_state.velocity * np.cos(delta_orientation) - ego_state.velocity
            p_rel_lead = distance_abs
            o_lead = obs_state
            obstacle_lead = obstacle

        return v_rel_follow, p_rel_follow, o_follow, obstacle_follow, v_rel_lead, p_rel_lead, o_lead, obstacle_lead

    def _get_rel_v_p_lane_based(
            self, obstacles_lanelet_ids: List[int], obstacle_states: List[State], lanelet_dict: Dict[str, List[int]],
            adj_obstacles: List[Obstacle]) -> Tuple[List[float], List[float], List[State], List[Obstacle], np.array]:

        """
        Get the relative velocity and position of obstacles in adj left, adj right and ego lanelet.
        In each lanelet, compute only the nearest leading and following obstacles.

        :param obstacles_lanelet_ids: The list of lanelets of obstacles
        :param obstacle_states: The list of states of obstacles
        :param lanelet_dict: The lanelet dictionary
            stores the list of lanelet ids by given keywords as (ego_all, ego_right....)
        :param adj_obstacles: lane-based adjacent obstacles
        :return: Relative velocities, relative positions, and detected obstacle states
        """
        # Initialize dummy values, in case no obstacles are present
        v_rel_left_follow, v_rel_same_follow, v_rel_right_follow, v_rel_left_lead, v_rel_same_lead, \
        v_rel_right_lead = [0.0] * 6

        p_rel_left_follow, p_rel_same_follow, p_rel_right_follow, p_rel_left_lead, p_rel_same_lead, \
        p_rel_right_lead = [self.max_obs_dist] * 6

        try:
            ego_vehicle_long_position, ego_vehicle_lat_position = self._local_ccosy.convert_to_curvilinear_coords(
                self._ego_state.position[0], self._ego_state.position[1])

            o_left_follow, o_left_lead, o_right_follow, o_right_lead, o_same_follow, o_same_lead = [None] * 6
            obstacle_left_follow, obstacle_left_lead, obstacle_right_follow, obstacle_right_lead, \
            obstacle_same_follow, obstacle_same_lead = [None] * 6

            for o_state, o_lanelet_id, obstacle in zip(obstacle_states, obstacles_lanelet_ids, adj_obstacles):

                distance_abs, distance_sign = self._get_ego_obstacle_distance(obstacle, o_state,
                                                                              (ego_vehicle_long_position,
                                                                               ego_vehicle_lat_position))

                if o_lanelet_id in lanelet_dict["ego_all"]:  # ego lanelet
                    v_rel_same_follow, p_rel_same_follow, o_same_follow, obstacle_same_follow, \
                    v_rel_same_lead, p_rel_same_lead, o_same_lead, obstacle_same_lead = \
                        self.get_rel_v_p_follow_leading(
                            distance_sign, distance_abs, p_rel_same_follow, p_rel_same_lead, v_rel_same_follow,
                            v_rel_same_lead, o_state, obstacle, self._ego_state, o_same_follow, o_same_lead,
                            obstacle_same_follow, obstacle_same_lead)

                if o_lanelet_id in lanelet_dict["right_all"]:  # right lanelet
                    v_rel_right_follow, p_rel_right_follow, o_right_follow, obstacle_right_follow, \
                    v_rel_right_lead, p_rel_right_lead, o_right_lead, obstacle_right_lead = \
                        self.get_rel_v_p_follow_leading(
                            distance_sign, distance_abs, p_rel_right_follow, p_rel_right_lead, v_rel_right_follow,
                            v_rel_right_lead, o_state, obstacle, self._ego_state, o_right_follow, o_right_lead,
                            obstacle_right_follow, obstacle_right_lead)

                if o_lanelet_id in lanelet_dict["left_all"]:  # left lanelet
                    v_rel_left_follow, p_rel_left_follow, o_left_follow, obstacle_left_follow, \
                    v_rel_left_lead, p_rel_left_lead, o_left_lead, obstacle_left_lead = \
                        self.get_rel_v_p_follow_leading(
                            distance_sign, distance_abs, p_rel_left_follow, p_rel_left_lead, v_rel_left_follow,
                            v_rel_left_lead, o_state, obstacle, self._ego_state, o_left_follow, o_left_lead,
                            obstacle_left_follow, obstacle_left_lead)

            detected_states = [o_left_follow, o_same_follow, o_right_follow, o_left_lead, o_same_lead, o_right_lead]
            detected_obstacles = [obstacle_left_follow, obstacle_same_follow, obstacle_right_follow,
                                  obstacle_left_lead, obstacle_same_lead, obstacle_right_lead]

        except ValueError:
            detected_states = [None] * 6
            detected_obstacles = [None] * 6
            ego_vehicle_lat_position = None

        v_rel = [v_rel_left_follow, v_rel_same_follow, v_rel_right_follow, v_rel_left_lead, v_rel_same_lead,
                 v_rel_right_lead]
        p_rel = [p_rel_left_follow, p_rel_same_follow, p_rel_right_follow, p_rel_left_lead, p_rel_same_lead,
                 p_rel_right_lead]

        return v_rel, p_rel, detected_states, detected_obstacles, ego_vehicle_lat_position

    def _get_ego_obstacle_distance(self, obstacle: Obstacle, obstacle_state: State, ego_curvi: Tuple[float, float]) \
            -> Tuple[float, int]:
        """
        Get the distance between the ego_vehicle and an obstacle

        :param obstacle: The obstacle in question
        :param ego_curvi: The position of the ego_vehicle in the curvi-system
        :returns: The absolute distance between the shapes
                / The absolute distance between the curvi_long positions depending on the configuration
                , The distance sign
        """

        ego_curvi_long_position, _ = ego_curvi

        try:
            o_curvi_long_position, _ = self._local_ccosy.convert_to_curvilinear_coords(obstacle_state.position[0],
                                                                                       obstacle_state.position[1])
        except ValueError:
            # the position is out of project area of curvilinear coordinate system
            o_curvi_long_position = ego_curvi_long_position + self.max_obs_dist

        distance_sign = np.sign(ego_curvi_long_position - o_curvi_long_position)

        if self.fast_distance_calculation:
            dist_abs = np.abs(ego_curvi_long_position - o_curvi_long_position)

        else:
            o_shape = obstacle.occupancy_at_time(self._current_time_step).shape.shapely_object
            dist_abs = self._ego_shape.distance(o_shape)

        return dist_abs, distance_sign

    @staticmethod
    def get_nearby_lanelet_id(connected_lanelet_dict: dict, ego_vehicle_lanelet: Lanelet) -> Tuple[dict, set]:
        """
        Get ids of nearby lanelets, e.g. lanelets that are successors, predecessors, left, or right of the
        `ego_vehicle_lanelet`
        additionally, all the connected lanelets of the nearby lanes are added. Connected lanelets are defined
        in the `connected_lanelet_dict`

        :param connected_lanelet_dict: A dict with its keys as lanelet id and values as connected lanelet ids
        :param ego_vehicle_lanelet: The list lanelets of the ego vehicle
        :return: A dict of nearby lanelets ids and the set of all nearby lanelets ids.
        """
        keys = {"ego", "left", "right", "ego_other", "left_other", "right_other", "ego_all", "left_all", "right_all", }
        lanelet_dict = {key: set() for key in keys}
        ego_vehicle_lanelet_id = ego_vehicle_lanelet.lanelet_id
        lanelet_dict["ego"].add(ego_vehicle_lanelet_id)  # Add ego lanelet

        for predecessor_lanelet_id in ego_vehicle_lanelet.predecessor:
            lanelet_dict["ego_other"].update(connected_lanelet_dict[predecessor_lanelet_id])
        for successor_lanelet_id in ego_vehicle_lanelet.successor:
            lanelet_dict["ego_other"].update(connected_lanelet_dict[successor_lanelet_id])

        if ego_vehicle_lanelet.adj_right_same_direction:
            # Get adj right lanelet with same direction
            lanelet_dict["right"].add(ego_vehicle_lanelet.adj_right)
        if ego_vehicle_lanelet.adj_left_same_direction:
            # Get adj left lanelet with same direction
            lanelet_dict["left"].add(ego_vehicle_lanelet.adj_left)

        for ego_lanelet_id in lanelet_dict["ego"]:
            lanelet_dict["ego_other"].update(connected_lanelet_dict[ego_lanelet_id])
        for left_lanelet_id in lanelet_dict["left"]:
            lanelet_dict["left_other"].update(connected_lanelet_dict[left_lanelet_id])
        for r in lanelet_dict["right"]:
            lanelet_dict["right_other"].update(connected_lanelet_dict[r])

        lanelet_dict["ego_all"] = set().union(set(lanelet_dict["ego"]), set(lanelet_dict["ego_other"]))

        lanelet_dict["left_all"] = set().union(set(lanelet_dict["left"]), set(lanelet_dict["left_other"]))
        lanelet_dict["right_all"] = set().union(set(lanelet_dict["right"]), set(lanelet_dict["right_other"]))

        all_lanelets_set = set().union(lanelet_dict["ego_all"], lanelet_dict["left_all"], lanelet_dict["right_all"])

        return lanelet_dict, all_lanelets_set


if __name__ == "__main__":
    import yaml
    from commonroad_rl.gym_commonroad.constants import PATH_PARAMS

    config_file = PATH_PARAMS["configs"]["commonroad-v1"]
    with open(config_file, "r") as config_file:
        config = yaml.safe_load(config_file)
    configs = config["env_configs"]
    surrounding_observation = SurroundingObservation(configs)
    print(surrounding_observation)



"""
Module EgoObservation
"""
from collections import defaultdict, OrderedDict
from typing import Union, Dict

import gym
import yaml
import numpy as np
from commonroad.common.solution import VehicleModel
from commonroad.scenario.lanelet import Lanelet
from commonroad.scenario.obstacle import State
from numpy import ndarray
from scipy import spatial

from commonroad_rl.gym_commonroad.constants import PATH_PARAMS
from commonroad_rl.gym_commonroad.action.vehicle import Vehicle, VehicleModel
from commonroad_rl.gym_commonroad.observation.observation import Observation
from commonroad_rl.gym_commonroad.utils.scenario import approx_orientation_vector, angle_difference


class EgoObservation(Observation):
    """
    Ego-vehicle-related observation class
    """

    def __init__(self, configs: Dict, configs_name: str = "ego_configs"):
        """

        :param configs: dictionary to store all observation configurations
        :param configs_name: key of configs dictionary corresponding to this observation
        """
        vehicle_model = configs["vehicle_params"]["vehicle_model"]
        self._is_PM_model = vehicle_model == VehicleModel.PM.value

        # Read config
        ego_configs = configs[configs_name]
        self.observe_v_ego: bool = ego_configs.get("observe_v_ego")
        self.observe_a_ego: bool = ego_configs.get("observe_a_ego")
        self.observe_jerk_ego: bool = ego_configs.get("observe_jerk_ego") and vehicle_model == VehicleModel.QP.value
        self.observe_relative_heading: bool = ego_configs.get("observe_relative_heading") and not self._is_PM_model
        self.observe_steering_angle: bool = ego_configs.get("observe_steering_angle") and not self._is_PM_model
        # yaw_rate
        self.observe_global_turn_rate: bool = ego_configs.get("observe_global_turn_rate") and not self._is_PM_model
        # TODO: add observation for jerk and slip_angle for linearized vehicle model (QP)
        # stores kappa for QP vehicle model
        self.observe_slip_angle: bool = ego_configs.get("observe_slip_angle") and vehicle_model == VehicleModel.QP.value
        self.observe_remaining_steps: bool = ego_configs.get("observe_remaining_steps")
        self.observe_is_friction_violation: bool = ego_configs.get("observe_is_friction_violation")

        self.observation_dict = OrderedDict()
        self.observation_history_dict = defaultdict(list)

    def build_observation_space(self) -> OrderedDict:
        observation_space_dict = OrderedDict()

        if self.observe_v_ego:
            if self._is_PM_model:
                observation_space_dict["v_ego"] = gym.spaces.Box(-np.inf, np.inf, (2,), dtype=np.float32)
            else:
                observation_space_dict["v_ego"] = gym.spaces.Box(-np.inf, np.inf, (1,), dtype=np.float32)
        if self.observe_a_ego:
            if self._is_PM_model:
                observation_space_dict["a_ego"] = gym.spaces.Box(-np.inf, np.inf, (2,), dtype=np.float32)
            else:
                observation_space_dict["a_ego"] = gym.spaces.Box(-np.inf, np.inf, (1,), dtype=np.float32)
        if self.observe_jerk_ego:
            observation_space_dict["jerk_ego"] = gym.spaces.Box(-np.inf, np.inf, (1,), dtype=np.float32)
        if self.observe_steering_angle:
            observation_space_dict["steering_angle"] = gym.spaces.Box(-np.pi, np.pi, (1,), dtype=np.float32)
        if self.observe_relative_heading:
            observation_space_dict["relative_heading"] = gym.spaces.Box(-np.pi, np.pi, (1,), dtype=np.float32)
        if self.observe_global_turn_rate:
            observation_space_dict["global_turn_rate"] = gym.spaces.Box(-np.inf, np.inf, (1,), dtype=np.float32)
        if self.observe_slip_angle:
            observation_space_dict["slip_angle"] = gym.spaces.Box(-np.inf, np.inf, (1,), dtype=np.float32)
        if self.observe_is_friction_violation:
            observation_space_dict["is_friction_violation"] = gym.spaces.Box(0, 1, (1,), dtype=np.int8)
        if self.observe_remaining_steps:
            observation_space_dict["remaining_steps"] = gym.spaces.Box(-np.inf, np.inf, (1,), dtype=np.float32)

        return observation_space_dict

    def observe(self, ego_lanelet: Lanelet, ego_vehicle: Vehicle, episode_length: int) -> Union[ndarray, Dict]:
        """
        Create ego-related observation for given state in an environment.
        """
        ego_state = ego_vehicle.state

        if self.observe_v_ego:
            if ego_vehicle.vehicle_model == VehicleModel.PM:
                self.observation_dict["v_ego"] = np.array([ego_state.velocity, ego_state.velocity_y])
            else:
                self.observation_dict["v_ego"] = np.array([ego_state.velocity])

        if self.observe_a_ego:
            if ego_vehicle.vehicle_model == VehicleModel.PM:
                self.observation_dict["a_ego"] = np.array([ego_state.acceleration, ego_state.acceleration_y])
            else:
                self.observation_dict["a_ego"] = np.array([ego_state.acceleration])

        if self.observe_jerk_ego:
            self.observation_dict["jerk_ego"] = np.array([ego_state.jerk])

        if self.observe_steering_angle:
            self.observation_dict["steering_angle"] = np.array([ego_state.steering_angle])

        if self.observe_relative_heading:
            relative_heading = EgoObservation.get_lane_relative_heading(ego_state, ego_lanelet)
            self.observation_dict["relative_heading"] = relative_heading

        if self.observe_global_turn_rate:
            self.observation_dict["global_turn_rate"] = np.array([ego_state.yaw_rate])

        if self.observe_slip_angle:
            self.observation_dict["slip_angle"] = np.array([ego_state.slip_angle])

        if self.observe_is_friction_violation:
            is_friction_violation = self._check_friction_violation(ego_vehicle)
            self.observation_dict["is_friction_violation"] = np.array([is_friction_violation])

        if self.observe_remaining_steps:
            self.observation_dict["remaining_steps"] = np.array([episode_length - ego_state.time_step])

        return self.observation_dict

    def draw(self, render_configs: Dict, **kwargs):
        """ Method to draw the observation """

    @staticmethod
    def _check_friction_violation(ego_vehicle: Vehicle):
        return ego_vehicle.violate_friction

    @staticmethod
    def get_lane_relative_heading(ego_vehicle_state: State, ego_vehicle_lanelet: Lanelet) -> float:
        """
        Get the heading angle in the Frenet frame.

        :param ego_vehicle_state: state of ego vehicle
        :param ego_vehicle_lanelet: lanelet of ego vehicle
        :return: heading angle in frenet coordinate system relative to lanelet center vertices between -pi and pi
        """
        lanelet_angle = EgoObservation._get_orientation_of_polyline(ego_vehicle_state.position,
                                                                    ego_vehicle_lanelet.center_vertices)

        return angle_difference(approx_orientation_vector(lanelet_angle),
                                approx_orientation_vector(ego_vehicle_state.orientation))

    @staticmethod
    def _get_orientation_of_polyline(position: np.array, polyline: np.array) -> float:
        """
        Get the approximate orientation of the lanelet.

        :param position: position to calculate the orientation of lanelet
        :param polyline: polyline to calculate the orientation of
        :return: orientation (rad) of the lanelet
        """
        # TODO: This method could make use of the general relative orientation function or could be fully replaced by
        #  the functions moved to the route planner

        idx = spatial.KDTree(polyline).query(position)[1]
        if idx < position.shape[0] - 1:
            orientation = np.arccos((polyline[idx + 1, 0] - polyline[idx, 0]) /
                                    np.linalg.norm(polyline[idx + 1] - polyline[idx]))
            sign = np.sign(polyline[idx + 1, 1] - polyline[idx, 1])
        else:
            orientation = np.arccos((polyline[idx, 0] - polyline[idx - 1, 0]) /
                                    np.linalg.norm(polyline[idx] - polyline[idx - 1]))
            sign = np.sign(polyline[idx, 1] - polyline[idx - 1, 1])
        if sign >= 0:
            orientation = np.abs(orientation)
        else:
            orientation = -np.abs(orientation)
        # orientation = shift_orientation(orientation)
        return orientation


if __name__ == "__main__":
    CONFIG_FILE = PATH_PARAMS["configs"]["commonroad-v1"]
    with open(CONFIG_FILE, "r") as config_file:
        config = yaml.safe_load(config_file)
    ego_observation = EgoObservation(config["env_configs"])
    print(ego_observation)



"""Abstract class for rewards"""

__author__ = "Xiao Wang, Brian Liao, Niels Muendler, Peter Kocsis, Armin Ettenhofer"
__copyright__ = "TUM Cyber-Physical System Group"
__credits__ = [""]
__version__ = "1.0"
__maintainer__ = "Xiao Wang"
__email__ = "xiao.wang@tum.de"
__status__ = "Released"

from abc import ABC, abstractmethod

from commonroad_rl.gym_commonroad.action import Action


class Reward(ABC):
    """Abstract class for rewards"""

    def reset(self, observation_dict: dict, ego_action: Action):
        pass

    @abstractmethod
    def calc_reward(self, observation_dict: dict, ego_action: Action) -> float:
        """
        Calculate the reward according to the observations

        :param observation_dict: current observations
        :param ego_action: Current ego_action of the environment
        :return: Reward of this step
        """



import numpy as np

from commonroad.geometry.shape import Rectangle
from commonroad.scenario.lanelet import Lanelet, LaneletNetwork, LaneletType
from commonroad_rl.gym_commonroad.action import Vehicle
from commonroad_rl.gym_commonroad.observation import ObservationCollector
from commonroad_rl.gym_commonroad.utils.scenario import approx_orientation_vector
from commonroad.geometry.shape import Shape
from commonroad.scenario.scenario import Scenario
from commonroad_dc.collision.collision_detection.pycrcc_collision_dispatch import create_collision_object
from commonroad.scenario.obstacle import ObstacleType, Obstacle, DynamicObstacle
from commonroad.prediction.prediction import Occupancy
import commonroad_dc.pycrcc as pycrcc


def check_collision_type(info, LOGGER, ego_vehicle: Vehicle, observation_collector: ObservationCollector,
                         scenario: Scenario, benchmark_id, lane_change_time_threshold, local_ccosy):
    other_vehicle_at_fault, reason = _current_collision_type_checking(
        ego_vehicle, observation_collector, scenario, lane_change_time_threshold, local_ccosy)
    if other_vehicle_at_fault:
        LOGGER.debug(f"Unexpected Collision not caused by ego vehicle in {benchmark_id} with reason: {reason}")

    info['valid_collision'] = not other_vehicle_at_fault
    info['collision_reason'] = reason


def _current_collision_type_checking(ego_vehicle: Vehicle, observation_collector: ObservationCollector,
                                     scenario: Scenario, lane_change_time_threshold, local_ccosy):

    # Determine collision participants
    collision_ego_vehicle = ego_vehicle.collision_object
    collision_checker = observation_collector.get_collision_checker()
    collision_objects = collision_checker.find_all_colliding_objects(collision_ego_vehicle)
    current_step = ego_vehicle.current_time_step

    # TODO: refactor once issue https://gitlab.lrz.de/cps/commonroad-drivability-checker/-/issues/17 is closed
    # get positions of all obstacles in the scenario
    obstacle_state_dict = dict()
    for obstacle in scenario.obstacles:
        obstacle_state = obstacle.state_at_time(current_step)
        if obstacle_state is not None:
            obstacle_state_dict[obstacle.obstacle_id] = obstacle_state

    collision_obstacles = {}
    for collision_object in collision_objects:
        collision_shape = collision_object.obstacle_at_time(current_step)
        if collision_shape.collide(observation_collector.road_edge["boundary_collision_object"]):
            return True, "other_outside_of_lanelet_network"
        for obstacle_id, obstacle_state in obstacle_state_dict.items():
            if isinstance(collision_shape, pycrcc.RectOBB) or isinstance(collision_shape, pycrcc.Circle):
                collision_shape_center = collision_shape.center()
            elif isinstance(collision_shape, pycrcc.RectAABB): # TODO: remove after dc adds center for RectAABB
                collision_shape_center = 0.5 * np.array([collision_shape.min_x() + collision_shape.max_x(),
                                                         collision_shape.min_y() + collision_shape.max_y()])
            else:
                raise NotImplementedError
            if np.allclose(collision_shape_center, obstacle_state.position):
                collision_obstacles[obstacle_id] = (obstacle_state, collision_shape)

    ego_shape = Rectangle(length=ego_vehicle.parameters.l,
                          width=ego_vehicle.parameters.w,
                          center=ego_vehicle.state.position,
                          orientation=ego_vehicle.state.orientation)
    ego_position = ego_vehicle.state.position

    # TODO: save lane results in pre-processing
    merge_results = create_lanes(scenario.lanelet_network, merging_length=1000.)
    lanes = [merge_result[1] for merge_result in merge_results]
    time_steps_back = int(lane_change_time_threshold / scenario.dt)

    for obstacle_id, (obstacle_state, collision_shape) in collision_obstacles.items():
        obstacle = scenario.obstacle_by_id(obstacle_id)
        # get edges of obstacle shape and check if it was fully on the lanelet network
        if current_step == obstacle.initial_state.time_step:
            return True, "other_suddenly_appear"
        if obstacle.obstacle_type == ObstacleType.CAR:
            # compute contact point of collision
            contact_point = _find_collision_point(ego_shape, obstacle.occupancy_at_time(current_step))
            lanelets_contact_point = set(scenario.lanelet_network.find_lanelet_by_position([np.array(contact_point)])[0])

            def find_lanelet_id_for_state(observation_collector, state, scenario):
                lanelet_polygons, lanelet_polygons_sg = observation_collector._get_lanelet_polygons(str(scenario.scenario_id))
                lanelet_ids = observation_collector.sorted_lanelets_by_state(
                    scenario, state, lanelet_polygons, lanelet_polygons_sg
                )
                return lanelet_ids[0]

            def find_lanelet_ids(state_list, time_steps):
                if len(state_list) >= time_steps:
                    state_list = state_list[-time_steps:]
                return set([find_lanelet_id_for_state(observation_collector, state, scenario)
                                      for state in state_list])

            def check_lane_change(lanelet_ids, lanes):
                for lane in lanes:
                    if (lanelet_ids.issubset(set(lane))):
                        return False
                return True

            obstacle_state_list = [obstacle.state_at_time(time_step) for time_step in
                                   range(obstacle.initial_state.time_step, current_step + 1)]

            ego_lanelet_ids = find_lanelet_ids(ego_vehicle.state_list, time_steps_back)
            obstacle_lanelet_ids = find_lanelet_ids(obstacle_state_list, time_steps_back)
            obstacle_exceeds_lane = True
            ego_exceeds_lane = True
            for lanelet_contact_point in lanelets_contact_point:
                if lanelet_contact_point in obstacle_lanelet_ids:
                    obstacle_exceeds_lane = False
                if lanelet_contact_point in ego_lanelet_ids:
                    ego_exceeds_lane = False
            if obstacle_exceeds_lane:
                return True, "other_out_of_lane"
            if ego_exceeds_lane:
                return False, "ego_out_of_lane"

            # determine back collision
            s_ego, _ = local_ccosy.convert_to_curvilinear_coords(ego_position[0], ego_position[1])
            s_obs, _ = local_ccosy.convert_to_curvilinear_coords(obstacle_state.position[0], obstacle_state.position[1])
            rear_collision = s_obs <= s_ego

            if rear_collision:
                # check if ego switched lane
                if check_lane_change(ego_lanelet_ids, lanes):
                    return False, "ego_cut_in"
                else:
                    return True, "other_collides_from_rear"
            else:
                # check if other switched lane
                if check_lane_change(obstacle_lanelet_ids, lanes):
                    return True, "other_cut_in"
                else:
                    return False, "ego_collides_from_rear"

    return False, "unsuccessful_determination"


def create_lanes(lanelet_network: LaneletNetwork, merging_length: float = 600.):
    """
    Creates lanes for road network

    :param merging_length: length for merging successors
    :param lanelet_network:
    :return:
    """
    lane_lanelets = []
    start_lanelets = []

    for lanelet in lanelet_network.lanelets:
        if len(lanelet.predecessor) == 0:
            start_lanelets.append(lanelet)
        else:
            predecessors = [lanelet_network.find_lanelet_by_id(pred_id) for pred_id in lanelet.predecessor]
            for pred in predecessors:
                if not lanelet.lanelet_type == pred.lanelet_type:
                    start_lanelets.append(lanelet)

    for lanelet in start_lanelets:
        merged_lanelets, merge_jobs = \
            Lanelet.all_lanelets_by_merging_successors_from_lanelet(
                lanelet, lanelet_network, merging_length)
        if len(merged_lanelets) == 0 or len(merge_jobs) == 0:
            merged_lanelets.append(lanelet)
            merge_jobs.append([lanelet.lanelet_id])
        for idx in range(len(merged_lanelets)):
            lane_lanelets.append((merged_lanelets[idx], merge_jobs[idx]))

    return lane_lanelets


def _find_collision_point(ego_shape: Shape, obstacle_occupancy: Occupancy, collision_checker: pycrcc.CollisionChecker=None):
    obstacle_vertices = obstacle_occupancy.shape.vertices

    # collision obstacle's corners in ego shape?
    for obstacle_vertice in obstacle_vertices:
        if ego_shape.contains_point(obstacle_vertice):
            return obstacle_vertice

    # ego's corner in obstacle's shape?
    # ego corners in collision obstacle's shape?
    for ego_vertice in ego_shape.vertices:
        if obstacle_occupancy.shape.contains_point(ego_vertice):
            return ego_vertice

    # neither of two cases
    if collision_checker is None:
        collision_checker = pycrcc.CollisionChecker()
        collision_checker.add_collision_object(create_collision_object(obstacle_occupancy.shape))
    ego_center = ego_shape.center
    for ego_vertice in ego_shape.vertices:
        intervals = collision_checker.raytrace(ego_center[0], ego_center[1], ego_vertice[0], ego_vertice[1], False)
        if len(intervals) > 0:
            return np.array([intervals[0][0], intervals[0][1]])

    raise ValueError


__author__ = "Peter Kocsis"
__copyright__ = "TUM Cyber-Physical System Group"
__credits__ = []
__version__ = "0.1"
__maintainer__ = "Peter Kocsis"
__email__ = "peter.kocsis@tum.de"
__status__ = "Integration"

"""
Test configuration script, contains the hooks of the pytest framework
"""

import shutil
from typing import Dict

from pytest import Function

from commonroad_rl.tests.common.marker import *
from commonroad_rl.tests.common.path import *

# Dictionary contains the items failed because of syntax error
_test_failed_syntax_error: Dict[Function, str] = {}


def pytest_addoption(parser, pluginmanager):
    """
    See https://docs.pytest.org/en/latest/reference.html#_pytest.hookspec.pytest_addoption
    """
    parser.addoption(
        "--scope",
        type=RunScope,
        nargs="+",
        default=list(RunScope),
        help="(tuple) The scope of the tests to be tested",
        choices=list(RunScope),
    )

    parser.addoption(
        "--type",
        type=RunType,
        nargs="+",
        default=list(RunType),
        help="(tuple) The scope of the tests to be tested",
        choices=list(RunType),
    )


def pytest_configure(config):
    """
    See https://docs.pytest.org/en/latest/reference.html#_pytest.hookspec.pytest_configure
    You can modify the configuration values here
    e.g: config.addinivalue_line("markers", "env(name): mark test to run only on named environment")
    """
    pass


def pytest_collection_modifyitems(items, config):
    """
    See https://docs.pytest.org/en/latest/reference.html#_pytest.hookspec.pytest_collection_modifyitems
    """
    # Select items
    selected_items = []
    deselected_items = []

    scopes = config.option.scope
    types = config.option.type
    for item in items:
        # The scope and type must be specified
        scope_marker = item.get_closest_marker("scope")
        type_marker = item.get_closest_marker("type")
        if scope_marker is None:
            _test_failed_syntax_error[
                item
            ] = "The scope of the test has not been defined"
            selected_items.append(item)
            continue
        if type_marker is None:
            _test_failed_syntax_error[
                item
            ] = "The type of the test has not been defined"
            selected_items.append(item)
            continue

        # Select the tests
        if scope_marker.args[0] in scopes and type_marker.args[0] in types:
            selected_items.append(item)
        else:
            deselected_items.append(item)

    # Apply the selection
    config.hook.pytest_deselected(items=deselected_items)
    items[:] = selected_items


def pytest_runtest_setup(item):
    """
    See https://docs.pytest.org/en/latest/reference.html#_pytest.hookspec.pytest_runtest_setup
    """
    item_output_path = output_root(item.fspath.purebasename)
    if os.path.exists(item_output_path) and os.path.isdir(item_output_path):
        shutil.rmtree(item_output_path)


def pytest_runtest_call(item):
    """
    See https://docs.pytest.org/en/latest/reference.html#_pytest.hookspec.pytest_runtest_call
    """
    print(item)
    if item in _test_failed_syntax_error:
        pytest.xfail(_test_failed_syntax_error[item])


def pytest_runtest_teardown(item):
    """
    See https://docs.pytest.org/en/latest/reference.html#_pytest.hookspec.pytest_runtest_teardown
    """
    pass


def pytest_runtest_makereport(item, call):
    """
    See https://docs.pytest.org/en/latest/reference.html#_pytest.hookspec.pytest_runtest_makereport
    """
    pass



"""
Module for CommonRoad Gym environment related constants
"""
# Lanelet parameters
from commonroad_rl.gym_commonroad.utils.scenario_io import get_project_root

# Visualization parameters
DRAW_PARAMS = {
    "draw_shape": True,
    "draw_icon": True,
    "draw_bounding_box": True,
    "trajectory_steps": 2,
    "show_label": False,
    "occupancy": {
        "draw_occupancies": 0,
        "shape": {
            "rectangle": {
                "opacity": 0.2,
                "facecolor": "#fa0200",
                "edgecolor": "#0066cc",
                "linewidth": 0.5,
                "zorder": 18,
            }
        },
    },
    "shape": {
        "rectangle": {
            "opacity": 1.0,
            "facecolor": "#fa0200",
            "edgecolor": "#831d20",
            "linewidth": 0.5,
            "zorder": 20,
        }
    },
}

# Path
ROOT_STR = str(get_project_root())

PATH_PARAMS = {
    "visualization": ROOT_STR + "/img",
    "pickles": ROOT_STR + "/pickles",
    "meta_scenario": ROOT_STR + "/pickles/meta_scenario",
    "train_reset_config": ROOT_STR + "/pickles/problem_train",
    "test_reset_config": ROOT_STR + "/pickles/problem_test",
    "log": ROOT_STR + "/log",
    "commonroad_solution": ROOT_STR + "/cr_solution",
    "configs": {"commonroad-v1": ROOT_STR + "/commonroad_rl/gym_commonroad/configs.yaml",
                "cr_monitor-v0": ROOT_STR + "/commonroad_rl/gym_commonroad/configs.yaml"}
}



"""File containing a reward function generator (Can't be incorporated into other modules because of cyclic import
problems """
from commonroad_rl.gym_commonroad.reward.hybrid_reward import HybridReward
from commonroad_rl.gym_commonroad.reward.reward import Reward
from commonroad_rl.gym_commonroad.reward.sparse_reward import SparseReward

reward_type_to_class = {
    "sparse_reward": SparseReward,
    "hybrid_reward": HybridReward,
}


def make_reward(configs: dict) -> Reward:
    """
    Initializes the reward class according to the env_configurations

    :param configs: The configuration of the environment
    :return: Reward class, either hybrid, sparse or dense
    """

    reward_type = configs["reward_type"]

    if reward_type in reward_type_to_class:
        return reward_type_to_class[reward_type](configs)
    else:
        raise ValueError(f"Illegal reward type: {reward_type}!")



"""
A tool to manipulate scenarios (rotate and translate) and save the modified version.
"""

__author__ = "Michael Feil"
__copyright__ = "TUM Cyber-Physical System Group"
__credits__ = [""]
__version__ = "1.0"
__maintainer__ = "Michael Feil"
__email__ = "michael.feil@tum.de"
__status__ = "Released"

import argparse
import logging
import os

import numpy as np
from typing import Tuple, List
import glob
import copy
import multiprocessing

from commonroad.planning.planning_problem import PlanningProblemSet
from commonroad.scenario.scenario import Scenario
from commonroad.common.file_reader import CommonRoadFileReader
from commonroad.common.file_writer import CommonRoadFileWriter, OverwriteExistingFile

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.DEBUG)
handler = logging.StreamHandler()
handler.setLevel(logging.DEBUG)
LOGGER.addHandler(handler)


def get_args() -> Tuple[argparse.ArgumentParser, np.ndarray]:
    """Scan arguments"""
    parser = argparse.ArgumentParser(
        description="Analyzes goal_definitions of scenarios and configures goal_observations for model",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--problem_dir_in",
        "-i",
        help="Path to input xml",
        type=str,
        default="/home/rl_students_ss21/data/inD/xml",
    )
    parser.add_argument(
        "--problem_dir_out",
        "-o",
        help="Path to folder of input xml",
        default="/home/rl_students_ss21/data/inD/rotated_xml",
    )
    parser.add_argument(
        "--rotations",
        "-r",
        help="list of rotations [0, 2Pi]",
        nargs="+",
        type=float,
        default=[np.pi / 2],
    )
    parser.add_argument(
        "--translations_x",
        "-tx",
        help="list of translations in x",
        nargs="+",
        type=float,
        default=[0],
    )
    parser.add_argument(
        "--translations_y",
        "-ty",
        help="list of translations in y",
        nargs="+",
        type=float,
        default=[0],
    )
    args = parser.parse_args()

    rot = np.asarray(args.rotations)
    tx = np.asarray(args.translations_x)
    ty = np.asarray(args.translations_y)

    if not (len(rot) == len(tx) == len(ty)):
        raise Exception(
            "length of rotations, translations_x, translations_y must be equal"
        )
    for i, angle in enumerate(rot):
        if angle > 2 * np.pi or angle < 0:
            LOGGER.warning(f"angle {i} is outside [0, {2*np.pi}]: {angle}")
    # check no angle is duplicated, i.e. no scenario is compute twice the same way.
    rot = rot % (2 * np.pi)
    rot_tx_ty = np.vstack((rot, tx, ty))
    unique = np.unique(rot_tx_ty, axis=1)
    if not np.array_equal(rot_tx_ty, unique):
        raise Exception(
            f"each set of (rotations, translations_x, translations_y) must be unique: \n {unique} \n {rot_tx_ty}"
        )
    args.rot = rot.tolist()
    return args, rot_tx_ty


def main() -> None:
    """
    Run rotation/translation of xmls with imports goal-requirements of scenarios and sets goal-observations in config
    argparse arguemnts
        --problem_dir_in
        --problem_dir_out
        optinal for rotation/translation also a three list of arguments:
            --translations_x
            --translations_y
            --rotations
    """
    args, rot_tx_ty = get_args()

    # Check arguments
    assert os.path.isdir(
        args.problem_dir_in
    ), f"The problem directory doesn't exist {args.problem_dir_in}"
    os.makedirs(args.problem_dir_out, exist_ok=True)
    xml_scenarios = glob.glob(args.problem_dir_in + "**/*.xml")

    # split on workers
    no_splits = 1 if len(xml_scenarios) < 100 else multiprocessing.cpu_count()

    with multiprocessing.Pool(processes=no_splits) as pool:
        # each process uses a set of files
        pool.starmap(
            rotate_scenarios,
            [
                (xml_scenario_paths, rot_tx_ty, args.problem_dir_out)
                for xml_scenario_paths in np.array_split(xml_scenarios, no_splits)
            ],
        )


def rotate_scenarios(
    xml_scenario_paths: List[str], rot_tx_ty: np.ndarray, outdir: str
) -> None:
    """
    Run rotation/translation of xmls with imports goal-requirements of scenarios and sets goal-observations in config

    :param xml_scenario_paths: List of filepaths
    :param outdir: str, directory for output
    :param rot_tx_ty: np.ndarray, array of shape (n,3) with shape[1] is [rotation in rad, translation_x in m, translation_y in m],

    :return: None
    """
    try:
        print(
            f"start rotating {len(xml_scenario_paths)} sceanrios on {rot_tx_ty.shape[1]} settings"
        )
        for xml_scenario_path in xml_scenario_paths:
            # open scenario
            reader = CommonRoadFileReader(xml_scenario_path)
            scenario, planning_problem_set = reader.open()

            if len(rot_tx_ty.T) == 1:
                write_rotate_and_translate_scenario(
                    scenario,
                    planning_problem_set,
                    reader,
                    outdir,
                    np.squeeze(rot_tx_ty),
                    use_copy=False,
                )
            else:
                for rot_tx_ty_setting in rot_tx_ty.T:
                    print(rot_tx_ty_setting)
                    write_rotate_and_translate_scenario(
                        scenario,
                        planning_problem_set,
                        reader,
                        outdir,
                        rot_tx_ty_setting,
                    )
    except Exception as ex:
        LOGGER.exception(ex)


def write_rotate_and_translate_scenario(
    scenario: Scenario,
    planning_problem_set: PlanningProblemSet,
    reader: CommonRoadFileReader,
    outdir: str,
    rot_tx_ty_setting: List[float, float, float],
    use_copy=True,
) -> None:
    """
    creates rotated/translated xml for scenario and planning_problem_set in the folder outdir.
    Rotated by rot_tx_ty_setting[0]
    Translated by rot_tx_ty_setting[1:]

    :param scenario: Scenario to be rotated/translated
    :param planning_problem_set: PlanningProblemSet to be rotated/translated
    :param reader: CommonRoadFileReader, Instance
    :param outdir: str, directory for output
    :param rot_tx_ty_setting: List[rotation in rad, translation_x in m, translation_y in m],
    :param use_copy: default True, set to True if scenario used multiple times

    :return: None
    """
    if use_copy:
        scenario = copy.deepcopy(scenario)
        planning_problem_set = copy.deepcopy(planning_problem_set)
    scenario.translate_rotate(rot_tx_ty_setting[1:], rot_tx_ty_setting[0])
    planning_problem_set.translate_rotate(rot_tx_ty_setting[1:], rot_tx_ty_setting[0])

    writer = CommonRoadFileWriter(
        scenario,
        planning_problem_set,
        author=reader._get_author(),
        affiliation=reader._get_affiliation(),
        source=reader._get_source(),
    )

    new_filename = os.path.join(
        outdir,
        f"{scenario.benchmark_id}_{rot_tx_ty_setting[0]}_{rot_tx_ty_setting[1]}_{rot_tx_ty_setting[2]}.xml",
    )

    writer.write_to_file(
        filename=new_filename, overwrite_existing_file=OverwriteExistingFile.ALWAYS
    )


if __name__ == "__main__":
    main()
